package rwd.abstractsyntax.spiderdiagram;

import java.util.Vector;

/**
 * @author jaf28
 * A compound diagram is built from Diagrams using the composite pattern.
 *
 * I need to find out where other classes externally change the
 * diagrams attribute - should we allow this, or add new methods here to
 * alter the contents of the diagrams attribute?
 * (change with care...)
 * Who calls setChildren and why?
 * 
 */
public abstract class CompoundDiagram extends Diagram {
    protected Vector diagrams;

    CompoundDiagram(Vector components, String name) {
        super(name);
        diagrams = new Vector();
        for (int i = 0; i < components.size(); i++) {
            Diagram d = (Diagram) components.get(i);
            diagrams.add(d.getClone());
        }
        //this.renameOutside(new Vector());
    }

    /**
     * @see rwd.abstractsyntax.spiderdiagram.Diagram#setName(java.lang.String)
     * 
     * If the name of a compound diagram is set (changed) then an iteration
     * is performed to make sure each component diagram has a unique name,
     * Renaming occurs according to the renameOutside method.
     */
    public void setName(String newName) {
        super.setName(newName);
        renameOutside(new Vector());
    }

    public void setChildren(Vector a) {
        diagrams = a;
    }

    public int countChildren() {
        return diagrams.size();
    }

    /**
     * @see rwd.abstractsyntax.spiderdiagram.Diagram#findPartCalled(java.lang.String)
     * Uses the unique naming property to iterate through descendents to
     * find a Diagram with given name.
     * 
     * It returns an Vector of the parent (perhaps null) and the 
     * appropriately named part.
     */
    public Vector findPartCalled(String name) // returns parent,part
    {
        //System.out.println("looking for part called " + name);
        //System.out.println("in this : "+DiagramFileUser.toSaveString(this));
        Vector result = new Vector();
        //System.out.println("see whether \n"+this+" \nis like \n"+d);
        if (this.getName().equals(name)) {
            result.add(null); // parent
            result.add(this); // part
            //System.out.println("found sought diagram at top!");
            return result;
        }
        for (int i = 0; i < diagrams.size(); i++) {
            Diagram d = (Diagram) (diagrams.get(i));
            //System.out.println("look for part in child : "+d.hashCode());
            result = d.findPartCalled(name);
            if (result!=null && result.get(1) != null) {
                // part can't be null for a successful result
                // System.out.println("found sought diagram in d1!");
                //System.out.println("found part in child part is : "+result.get(1).hashCode());
                return result;
            }
        }
        return null;
    }

    public boolean isCompound() {
        return true;
    }

    public boolean isFalse() {
        return false;
    }

    public Vector getAllNames() {
        Vector result = new Vector();
        for (int i = 0; i < diagrams.size(); i++) {
            Diagram d = (Diagram) (diagrams.get(i));
            result.addAll(d.getAllNames());
        }
        return result;
    }

    public void renameOutside(Vector bannedNames) { //
        //System.out.println("compound: rename "+getName()+" outside "+bannedNames);
        while (bannedNames.contains(getName())) {
            //System.out.println("extending name in CompoundDiagram.renameOutside()");
            setName(Utilities.makeNameLike(getName()));
        }
        bannedNames.add(getName());
        //System.out.println("renamed this as "+getName()+" and rename subdiagrams with ban on "+bannedNames);
        for (int i = 0; i < diagrams.size(); i++) {
            Diagram d = (Diagram) (diagrams.get(i));
            d.renameOutside(bannedNames);
        }
    }

    public Vector getChildren() {
        return diagrams;
    }

    /**
     * @see rwd.abstractsyntax.spiderdiagram.Diagram#addUnitaryComponents(java.util.Vector)
     * used to collect together unitary components (leaves) of the heirarchy.
     */
    protected void addUnitaryComponents(Vector componentsSoFar) {
        Vector children = getChildren();
        for (int i = 0; i < children.size(); i++) {
            Diagram child = (Diagram) children.get(i);
            //System.out.println("child is "+child);
            //System.out.println("componentsSoFar are "+componentsSoFar);
            child.addUnitaryComponents(componentsSoFar);
        }
    }

    /**
     * A utility method to collect together the leaves of the tree.
     */
    public Vector getUnitaryParts() {
        Vector result = new Vector();
        addUnitaryComponents(result);
        return result;
    }
    /**
     * @see rwd.abstractsyntax.spiderdiagram.Diagram#getParts()
     * 
     */
    public Vector getParts() {
        Vector result = new Vector();
        result.add(this);
        for (int i = 0; i < diagrams.size(); i++) {
            Diagram child = (Diagram) diagrams.get(i);
            if (child instanceof UnitaryDiagram) {
                result.add(child);
            } else {
                result.addAll(((CompoundDiagram) child).getParts());
            }
        }
        return result;
    }

    public CompoundDiagram getParentOf(Diagram part) {
        //System.out.println("asked to find parent of "+part.toDebugString());
        //System.out.println("in "+this.toDebugString());
        int[] coordsOfPart = getCoords(part);
        if (coordsOfPart.length == 0) {
            return null;
        }
        int[] reducedCoords = new int[coordsOfPart.length - 1];
        for (int i = 0; i < reducedCoords.length; i++) {
            reducedCoords[i] = coordsOfPart[i];
        }
        return (CompoundDiagram) getPartAt(reducedCoords);
    }

    public int[] getCoords(Diagram part) {
        if (this == part) {
            return new int[] {
            };
        } else {
            for (int i = 0; i < getChildren().size(); i++) {
                Diagram child = (Diagram) getChildren().get(i);
                int[] coordsFromChild = child.getCoords(part);
                if (coordsFromChild != null) {
                    // add "i" on at start and return
                    int[] result = new int[coordsFromChild.length + 1];
                    result[0] = i;
                    for (int j = 0; j < coordsFromChild.length; j++) {
                        result[j + 1] = coordsFromChild[j];
                    }
                    return result;
                }
            }
            return null;
        }
    }

    public Diagram getPartAt(int[] coords) { //System.out.println("in compound diagram "+this.toDebugString());
        //System.out.println("looking for part at ");
        //for(int j=0; j<coords.length; j++)
        //{ /**/System.out.print(""+coords[j]); }
        //System.out.println("");
        if (coords.length == 0) {
            return this;
        }
        if (coords[0] > getChildren().size()) {
            return null;
        } else {
            int[] reducedCoords = new int[coords.length - 1];
            for (int i = 0; i < reducedCoords.length; i++) {
                reducedCoords[i] = coords[i + 1];
            }
            Diagram child = (Diagram) getChildren().get(coords[0]);
            //System.out.println("in child "+child.toDebugString());
            //System.out.println("look for part at ");
            //for(int j=0; j<reducedCoords.length; j++)
            //{ /**/System.out.print(""+reducedCoords[j]); }
            //System.out.println("");
            return child.getPartAt(reducedCoords);
        }
    }
}
