package rwd.abstractsyntax.spiderdiagram;

import java.util.List;
import java.util.Vector;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.io.Serializable;

import rwd.abstractsyntax.constraintdiagram.Arrow;
import rwd.abstractsyntax.eulerdiagram.*;
import rwd.concretesyntax.DiagramItem;
import rwd.ocl.constraintdiagram.DependenceGraph;
import rwd.ocl.constraintdiagram.Graph;
import rwd.ocl.constraintdiagram.ReadingTree;
import rwd.ocl.constraintdiagram.ReadingTreeNode;
import uk.ac.kent.cs.kmf.util.ILog;

/**
 * 
 * @author RWD,Richard Bosworth
 * 
 * @version 8.9.05   14:20
 */
public class UnitaryDiagram extends Diagram implements Serializable {
    private Vector zones = new Vector();
    private Vector shadedZones = new Vector();
    private Vector contours = new Vector();
    private Vector spiders = new Vector();
    private Vector arrows = new Vector();
    private ReadingTree readingTree = new ReadingTree("");

    protected UnitaryDiagram(String name) {
        super(name);
    }

    public UnitaryDiagram(Vector zones, String name) {
        super(name);
        this.zones.addAll(zones);
        for(int i=0; i<zones.size(); i++) {
            Zone z = (Zone) (zones.get(i));
            contours = z.getAllContours();
            if (contours.size() != 0)
            	break;
        }
    }
    public UnitaryDiagram() { // just for reading from file
    }
    public Vector getCloneOfContourSet() {
        return (Vector)contours.clone();
    }
    public Vector getCloneOfZoneSet() {
        return (Vector)zones.clone();
    }
    public boolean containsShaded(Zone z){
        return shadedZones.contains(z);
    }
    public boolean containsShadedAll(Collection c){
        return shadedZones.containsAll(c);
    }
    public Vector getSpiders() {
        return spiders;
    }
    
    /**
     * 
     * @param s name of required spider
     * @return spider with given name, or null if none
     */
    /*
     * test:
     *   diagram with no spiders
     *   diagram with two spiders, one of which is s
     *   diagram with two spiders, neither of which is s
     */
    public Spider getSpiderCalled(String s) {
    	Iterator sit = getSpiders().iterator();
    	while (sit.hasNext()) {
    		Spider sp = (Spider) sit.next();
    		if (sp.label.equals(s)) 
    			return sp;    		
    	}
    	return null;
    }
    
    public void addSpider(Vector habitat) {
        if (habitat.size() == 0) {
            throw new IllegalArgumentException();
        }
        Spider s = new Spider((Vector) (habitat.clone()));
        spiders.add(s);
    }
    public void addSpider(Spider s) {
        spiders.add(s);
    }
    public void removeSpider(Spider s) {
        spiders.remove(s);
    }
    public void addZone(Zone z){
        Vector zc = z.getAllContours();
        if(!zc.containsAll(contours) || ! contours.containsAll(zc)){
            System.out.println("trying to add the zone "+z+" to");
            System.out.println(""+DiagramFileUser.toSaveString(this));
            throw new IllegalArgumentException();            
        }
        if(zones.contains(z)){
            throw new IllegalArgumentException();            
        }
        zones.add(z);
    }
    public void removeZone(Zone z){
        if(!zones.contains(z)){
            throw new IllegalArgumentException();            
        }
        zones.remove(z);
        shadedZones.remove(z);
    }
    public Iterator getZoneIterator(){
        return zones.iterator();
    }
    public int getNumZones(){
        return zones.size();
    }
    public void shadeRegion(Vector zones) {
        //
        // only add zones not already present
        //
        for(int i=0; i<zones.size(); i++){
            Zone z = (Zone)zones.get(i);
            if(!shadedZones.contains(z)){
                shadedZones.add(z);
            } else{
                System.out.println("zones parameter was "+zones);
                throw new IllegalArgumentException();
            }
        }
    }
    public void shadeZone(Zone z) {
        if(shadedZones.contains(z)){
            throw new IllegalArgumentException();
        }
        shadedZones.add(z);
    }
    public Vector getArrows() {
        return arrows;
    }
    public void addArrow(Arrow arrow) {
        arrows.add(arrow);
    }
    public void removeArrow(Arrow arrow) {
        spiders.remove(arrow);
    }
    public boolean equals(Diagram d) {
        if (this instanceof FalseDiagram) {
            return (d instanceof FalseDiagram);
        } else if (d instanceof FalseDiagram) {
            return false;
        }
        //System.out.println("comparing unitary diagrams::::");
        //System.out.println("this  "+this);
        UnitaryDiagram otherd = null;
        if (d instanceof UnitaryDiagram) {
            otherd = (UnitaryDiagram) d;
        } else { //System.out.println("other is not a unitary diagram");
            return false;
        }
        //System.out.println("otherd"+otherd);
        //System.out.println(":::comparing uintary diagrams::::");
        // see if the contour sets are the same size
        //System.out.println("this.contours are "+this.contours);
        //System.out.println("otherd.contours are "+otherd.contours);
        Vector theOtherContours = otherd.contours;
        boolean b1 = contours.containsAll(theOtherContours);
        boolean b2 = theOtherContours.containsAll(contours);
        if (!b1 || !b2) {
            //System.out.println("contours don't match");
            return false;
        }
        // check the zone sets are same size
        if (!otherd.zones.containsAll(this.zones)
            || !this.zones.containsAll(otherd.zones)) {
            //System.out.println("zones don't match");
            return false;
        }
        // now we know underlying euler diagram matches
        // check the shaded zone sets are same size
        if (!otherd.shadedZones.containsAll(this.shadedZones)
            || !this.shadedZones.containsAll(otherd.shadedZones)) {
            //System.out.println("shaded zones don't match");
            return false;
        }
        // now we know underlying euler diagram with shading matches
        // check the spider sets are the same size
        if (otherd.spiders.size() != this.spiders.size()) {
            //System.out.println("num spiders don't match");
            return false;
        }
        // check that the spiders have matching habitats...
        Vector habitats = new Vector();
        for (int i = 0; i < spiders.size(); i++) {
            habitats.add(((Spider) (spiders.get(i))).habitat);
        }
        Vector habitats2 = new Vector();
        for (int i = 0; i < otherd.spiders.size(); i++) {
            habitats2.add(((Spider) (otherd.spiders.get(i))).habitat);
        }
        if (!zoneBagsMatch(habitats, habitats2)) {
            //System.out.println("spiders don't match");
            return false;
        }
        //System.out.println("diagrams match");
        return true;
    }
    public String toString() { //return "name: "+getName()+"\ncontours: "+contours+"\nzones: "+zones+
        //   "\nshadedzones: "+shadedZones+"\nspiders: "+spiders;
        return getName();
    }
    /** Override to String */
    public String toDebugString() {
        String res = "UnitaryDiagram {\n";
        res += "\t label: " + getName() + "\n";
        res += "\t contours: " + getCloneOfContourSet() + "\n";
        res += "\t zones: " + getCloneOfZoneSet() + "\n";
        res += "\t shaded zones: " + getCloneOfShadedZones() + "\n";
        res += "\t feet: " + getSpiders() + "\n";
        res += "\t arrows: " + getArrows() + "\n";
        res += "\t reading tree: " + getReadingForest() + "\n";
        res += "}";
        return res;
    }
    public static boolean zoneBagsMatch(
        Vector bag1,
        Vector bag2) { //System.out.println("zonebags "+bag1+" and "+bag2);
        if (bag1.size() != bag2.size()) {
            return false;
        }
        // boolean flags which zones in set2 have been "matched"
        boolean[] matches = new boolean[bag1.size()];
        for (int i = 0; i < bag1.size(); i++) {
            matches[i] = false;
        }
        // check each zone here is there
        for (int i = 0; i < bag1.size(); i++) {
            Vector zoneSet1 = (Vector) (bag1.get(i));
            matchSearch : for (int j = 0; j < bag2.size(); j++) {
                Vector zoneSet2 = (Vector) (bag2.get(j));
                if (matches[j] == false) {
                    //System.out.println("comparing "+zoneSet1+" and "+zoneSet2);
                    if (zoneSet1.containsAll(zoneSet2)
                        && zoneSet2.containsAll(zoneSet1)) {
                        matches[j] = true;
                        break matchSearch;
                    }
                }
            }
        }
        //System.out.println("got matches");
        for (int i = 0;
            i < matches.length;
            i++) { //System.out.println(""+matches[i]);
            if (!matches[i]) {
                return false;
            }
        }
        return true;
    }
    public void setNewZoneSet(Vector zones) {
        // why is this needed? Used by MDF
        this.zones = zones;
        Vector shadingToRemove = ((Vector) (shadedZones.clone()));
        shadingToRemove.removeAll(zones);
        shadedZones.removeAll(shadingToRemove);
        Zone z = (Zone) (zones.get(0));
        contours = z.getAllContours();
        //System.out.println("zones is now "+zones);
    }

    public void setShadedZones(Vector zones) {
        LinkedHashSet set = new LinkedHashSet(zones);
        if(set.size()!=zones.size()){
            throw new IllegalArgumentException();
        }
        shadedZones = zones;
    }

	public void addShadedZones(Vector zones) {
		shadedZones.addAll(zones);
	}

    public boolean aSpiderTouches(Zone z) {
        for (int i = 0; i < spiders.size(); i++) {
            Spider s = (Spider) (spiders.get(i));
            if (s.habitat.contains(z)) {
                return true;
            }
        }
        //System.out.println("in "+spiders+" no spider touches "+z);
        return false;
    }
    
    /**
     * Note: this method clones the constituent spiders of this diagram
     * @return clone of this diagram
     */
    /*
     * test:
     *   diagram with no content
     *   diagram with two contours, two spiders, two arrows, reading forest
     */
    public Diagram getClone() { // clone zones, shadedzones
        Vector newZones = (Vector) (zones.clone());
        Vector newShadedZones = (Vector) (shadedZones.clone());
        UnitaryDiagram ud = new UnitaryDiagram(newZones, getName());
        ud.shadedZones = newShadedZones;
        for (int i = 0; i < spiders.size(); i++) {
            Spider s = (Spider) (spiders.get(i));
            ud.addSpider(new Spider(s.getLabel(),(Vector) s.getHabitat()));
        }
        Iterator arit = arrows.iterator();
        while (arit.hasNext()){
        	ud.addArrow((Arrow)arit.next());
        }
        ud.setReadingForest(this.getReadingForest());
        return ud;
    }
    
    public Vector findPartCalled(String name) // returns parent,part
    {
        Vector result = new Vector();
        result.add(null); // null parent
        //System.out.println("compare "+name+" and "+this.getName());
        //System.out.println("compare "+name.length()+" and "+this.getName().length());

        if (name.equals(this.getName())) {
            //System.out.println("match");
            result.add(this);
        } // this is "part"
        else {
            //System.out.println("no match");
            result.add(null);
        } // null part
        return result;
    }
    
    /* public Vector findPartLike(Diagram d)// returns parent,part
     { //System.out.println("seeking\n"+d+"\n in \n"+this);
       Vector result = new Vector();
       result.add(null);// null parent
       if(this.isEqualTo(d))
       { result.add(this); }// this is "part"
       else
       { result.add(null); } // null part
       //System.out.println("returning \n"+result);
       return result;
     }*/

    public Vector getAllNames() {
        Vector result = new Vector();
        result.add(getName());
        return result;
    }

    public void flatten() {
    }

    public void renameOutside(Vector bannedNames) {
        //System.out.println("rename "+getName()+" outside "+bannedNames);
        while (bannedNames.contains(getName())) {
            setName(Utilities.makeNameLike(getName()));
        }
        bannedNames.add(getName());
        //System.out.println("renamed as "+getName()+" leaving ban on "+bannedNames);
    }
    
    public static Vector allPossibleZones(Vector cs) {
        Vector contourSets = Utilities.powerSet(cs);
        Vector result = new Vector();
        for (int i = 0; i < contourSets.size(); i++) {
            Vector contoursCopy = (Vector) (cs.clone());
            Vector ins = (Vector) contourSets.get(i);
            contoursCopy.removeAll(ins);
            result.add(Zone.getZone(ins, contoursCopy));
        }
        return result;
    }
    
    public Vector getUnitaryParts() {
        Vector result = new Vector();
        result.add(this);
        return result;
    }
    
    public Vector getParts() {
        Vector result = new Vector();
        result.add(this);
        return result;
    }
    
    protected void addUnitaryComponents(Vector componentsSoFar) {
        componentsSoFar.add(this);
    }

    public CompoundDiagram getParentOf(Diagram part) {
        return null;
    }
    public int[] getCoords(Diagram part) {
        if (this == part) {
            return new int[] {
            };
        } else {
            return null;
        }
    }
    public Diagram getPartAt(int[] coords) {
        if (coords.length > 0) {
            return null;
        } else {
            return this;
        }
    }

    public boolean containsNegationOf(Diagram part) {
        return false;
    }

    public Iterator getSZIterator() {
        return shadedZones.iterator();        
    }
    public int getNumShadedZones(){
        return shadedZones.size();
    }
    public Vector getCloneOfShadedZones(){
        return (Vector)shadedZones.clone();
    }

    public void clearShadedZones() {
        shadedZones.clear();
    }

    public void unshade(Zone z) {
        shadedZones.remove(z);
    }

    /**
     * 
     * @return string representation of reading tree for this diagram
     */
    /*
     * test: once thru with non-null tree
     */
    public String getReadingForest() {
    	return readingTree.toString();
	}
    
    /**
     * Post: reading tree for this diagram has been set as specified 
     *         by readingForest
     * @param readingForest string representation of required reading tree
     */
	public void setReadingForest(String readingForest) {
		this.readingTree = new ReadingTree(readingForest);
	}
	
	/** Checks if the reading tree is valid */
	// Check the following conditions
	// 1. the nodes of RT(d) are the nodes of DG(d) together with the node PTC
	// 2. if n1 is adjancent to n2 in DG(d) then n1 and n2 lie in a common path 
	// in RT(d) directed from the root node PTC
	public boolean isValidTree() {
		return isValidTree(null);
	}
	public boolean isValidTree(ILog log) {
		return isValidTree(getReadingForest(), log);
	}
	public boolean isValidTree(String forest, ILog log) {
		// Compute the reading tree
		ReadingTree tree = new ReadingTree(forest);
		// Compute the dependence graph
		DependenceGraph graph = getGraph();
		// Check conditions
		boolean flag1 = checkNodes(tree, graph, log);
		boolean flag2 = checkEdges(tree, graph, log);			
		return flag1 && flag2;
	}
	protected boolean checkNodes(ReadingTree tree, DependenceGraph graph, ILog log) {
		// Print message
		printMessage(log, "Checking nodes ...");
		List treeNodes = tree.getNodeNames();
		List graphNodes = graph.getNodes();
		if (treeNodes.size() != graphNodes.size()) {
			printMessage(log, "Reading tree and dependence graph do not have the same nodes");
			return false;
		}
		for(int i=0; i<graphNodes.size(); i++) {
			Object gNode = graphNodes.get(i);
			String label = "";
			if (gNode instanceof DiagramItem) {
				label = ((DiagramItem)gNode).getLabel();
			}
			if (!treeNodes.contains(label)) {
				printMessage(log, "Cannot find node '"+label+"' in reading tree");
				return false;
			}
		}
		return true;
	}
	protected boolean checkEdges(ReadingTree tree, DependenceGraph graph, ILog log) {
		List graphEdges = graph.getEdges();
		for(int i=0; i<graphEdges.size(); i++) {
			Graph.Pair edge = (Graph.Pair)graphEdges.get(i);
			DiagramItem n1 = (DiagramItem)edge.first;
			DiagramItem n2 = (DiagramItem)edge.second;
			if (!checkEdge(tree, n1.getLabel(), n2.getLabel(), log))
				return false;
		}
		return true;
	}
	protected boolean checkEdge(ReadingTree tree, String label1, String label2, ILog log) {
		ReadingTreeNode n1 = tree.find(label1);
		ReadingTreeNode n2 = tree.find(label2);
		if (n1 == null) {
			printMessage(log, "Cannot find node '"+label1+"' in reading tree");
		}
		if (n2 == null) { 
			printMessage(log, "Cannot find node '"+label2+"' in reading tree");
		}
		if (tree.isAncestor(label1, label2) || tree.isAncestor(label2, label1))
			return true;
		printMessage(log, "Wrong edge ('"+label1+"', '"+label2+"')");
		return false;
	}
	protected void printMessage(ILog log, String msg) {
		if (log != null)
			log.reportMessage(msg);
	}
	
	public DependenceGraph getGraph() {
		return new DependenceGraph(this);
	}

	public List getZones() {
		return zones;
	}

	/*
	// test program
    // ------------

    private static void pr(String s) {
      System.out.println(s);
    }

    private static void hd(String s) {
      System.out.println("\n" + s + "\n");
    }

    private static void pg() {
      try {
        int i;
        System.out.print("next page? (CR)");
        do i = System.in.read(); while (i != '\n');
        System.out.println();
      }
      catch (Exception e) {System.out.println("raised exception : " + e);}
    }

    public static void main(String args[]) throws Exception {
    	/*
    	{
	    	hd("tests of getClone");
	    	UnitaryDiagram ud1 = new UnitaryDiagram("ud1");
	    	pr("ud1="+ud1);
	    	pr("ud1.zones="+ud1.zones);
	    	pr("ud1.shadedZones="+ud1.shadedZones);
	    	pr("ud1.contours="+ud1.contours);
	    	pr("ud1.spiders="+ud1.spiders);
	    	pr("ud1.arrows="+ud1.arrows);
	    	pr("ud1.readingForest="+ud1.readingForest);
	    	pg();
	    	
	    	UnitaryDiagram ud1clone = (UnitaryDiagram) ud1.getClone();
	    	pr("ud1clone)="+ud1clone);
	    	pr("ud1clone.zones="+ud1clone.zones);
	    	pr("ud1clone.shadedZones="+ud1clone.shadedZones);
	    	pr("ud1clone.contours="+ud1clone.contours);
	    	pr("ud1clone.spiders="+ud1clone.spiders);
	    	pr("ud1clone.arrows="+ud1clone.arrows);
	    	pr("ud1clone.readingForest="+ud1clone.readingForest);
	     	pg();
	    	
	    	pr("make ud2 with contours a and b, zones x,y,z, spiders (x,y),(y,z), arrows (a,b),(b,a), forest (s1)(s2)");
	    	Contour a = Contour.getContour("a");
	    	Contour b = Contour.getContour("b");
	    	
	    
	    	Vector v = new Vector();
	    	Vector vab = new Vector();
	    	vab.add(a);
	    	vab.add(b);
	    	Zone x = Zone.getZone(v,vab);
	    	Vector va = new Vector();
	    	va.add(a);
	    	Vector vb = new Vector();
	    	vb.add(b);
	    	Zone y = Zone.getZone(va,vb);
	    	Zone z = Zone.getZone(vab,v);
	    	Vector vxyz = new Vector();
	    	vxyz.add(x);
	    	vxyz.add(y);
	    	vxyz.add(z);
	    	UnitaryDiagram ud2 = new UnitaryDiagram(vxyz,"ud2");
	    	
	    	Vector vxy = new Vector();
	    	vxy.add(x);
	    	vxy.add(y);
	    	Spider s1 = new Spider("s1",vxy);
	    	Vector vyz = new Vector();
	    	vyz.add(y);
	    	vyz.add(z);
	    	Spider s2 = new Spider("s2",vyz);
	    	ud2.addSpider(s1);
	    	ud2.addSpider(s2);
	    	
	    	Arrow ar1 = new Arrow(a,b,"a1");
	    	Arrow ar2 = new Arrow(b,a,"a2");
	    	ud2.addArrow(ar1);
	    	ud2.addArrow(ar2);
	    	
	    	ud2.setReadingForest("(s1)(s2)");
	    	
	    	pr("ud2="+ud2);
	    	pr("ud2.zones="+ud2.zones);
	    	pr("ud2.shadedZones="+ud2.shadedZones);
	    	pr("ud2.contours="+ud2.contours);
	    	pr("ud2.spiders="+ud2.spiders);
	    	pr("ud2.arrows="+ud2.arrows);
	    	pr("ud2.readingForest="+ud2.readingForest);
	     	pg();
	    	
	    	UnitaryDiagram ud2clone = (UnitaryDiagram) ud2.getClone();
	    	pr("ud2clone)="+ud2clone);
	    	pr("ud2clone.zones="+ud2clone.zones);
	    	pr("ud2clone.shadedZones="+ud2clone.shadedZones);
	    	pr("ud2clone.contours="+ud2clone.contours);
	    	pr("ud2clone.spiders="+ud2clone.spiders);
	    	pr("ud2clone.arrows="+ud2clone.arrows);
	    	pr("ud2clone.readingForest="+ud2clone.readingForest);
	     	pg();
    	}
    	
    	{
    		hd("Tests of getSpiderCalled");
	    	UnitaryDiagram ud = new UnitaryDiagram("ud");
	    	pr("ud"+ud.toDebugString());
	    	pr("ud.getSpiderCalled(\"s1\")="+ud.getSpiderCalled("s1"));
	    	pg();
	    	
	    	Spider s1 = new Spider("s1",null);
	    	Spider s2 = new Spider("s2",null);
	    	Spider s3 = new Spider("s3",null);
	    	ud.addSpider(s1);
	    	ud.addSpider(s2);
	    	pr("ud"+ud.toDebugString());
	    	pr("ud.getSpiderCalled(\"s1\")="+ud.getSpiderCalled("s1"));
	    	pr("ud.getSpiderCalled(\"s3\")="+ud.getSpiderCalled("s3"));
	    	pg();
    	}
    	
    	{
    		hd("Tests of getReadingForest, setReadingForest");
	    	UnitaryDiagram ud = new UnitaryDiagram("ud");
	    	pr("ud"+ud.toDebugString());
	    	pr("ud.getReadingForest()="+ud.getReadingForest());
	    	pg();
	    	
	    	pr("ud.setReadingForest(\"(a(b(c))(d(e)))\")");
	    	ud.setReadingForest("(a(b(c))(d(e)))");
	    	pr("ud"+ud.toDebugString());
	    	pr("ud.getReadingForest()="+ud.getReadingForest());
	    	pg();
	    	
	    	
    	}
    }
    */
}
