package rwd.concretesyntax.spiderdiagram;

import java.awt.*;
import java.awt.geom.*;

import rwd.concretesyntax.*;
import rwd.gui.editor.DiagramCanvas;

/**
 * This is something of a misnomer - it actually represents a connector
 * in a concrete diagram.
 * The connector may be a true spider leg, or a strand between two spiders,
 * or a tie between two spiders.
 * 
 * @author  op, rnb
 * @version 6.7.05  14:44
 */
public class ConcreteSpiderLeg 
	extends DiagramItem 
	implements java.io.Serializable 
{    
    public static final int NORMAL = 1;
    public static final int STRAND = 2;
    public static final int TIE = 3;
    
    public ConcreteSpiderLeg(String label, ConcreteSpiderFoot source, ConcreteSpiderFoot target, int type) {
    	_label = label;
        _source = source;
        _target = target;
        this.setType(type);
        this.setBackColor(Color.BLACK);
    }

	public void draw(Graphics2D g, boolean smoothed) {
		Shape shape = getShape(smoothed);
		getSource().draw(g, smoothed);
		getTarget().draw(g, smoothed);
		BasicStroke stroke = new BasicStroke(
				getThickness(),
				BasicStroke.CAP_BUTT,
				BasicStroke.JOIN_ROUND);
		g.setStroke(stroke);
		g.setColor(getForeColor());
		g.draw(shape);		
	}

    public Shape getShape(boolean smooth) {
    	Shape sourceShape = _source.getShape();
    	Shape targetShape = _target.getShape();
        Shape shape = null;
        GeneralPath path = null;
        int x1 = _source.getX();
        int y1 = _source.getY();
        int x2 = _target.getX();
        int y2 = _target.getY();
        switch (_type) {
            case NORMAL:
            	path = new GeneralPath();
            	path.moveTo(x1, y1);
            	path.lineTo(x2, y2);
				path.quadTo((x1 + x2) / 2, (y1 + y2) / 2, x2, y2);
//				// Add feet
//				path.append(sourceShape, false);
//				path.append(targetShape, false);
                shape = path;
                break;
            case STRAND:
                int xmid = (x1 + x2) / 2;
                int ymid = (y1 + y2) / 2;
                int xmidleft = xmid - (ymid - y1);
                int ymidleft = ymid + (xmid - x1);
                int xmidright = xmid + (ymid - y1);
                int ymidright = ymid - (xmid - x1);
                path = new GeneralPath();
                path.moveTo(x1, y1);
                //path.lineTo(xmidleft, ymidleft);
                //path.lineTo(xmidright, ymidright);
                //path.lineTo(x2, y2);
                path.curveTo(xmidleft, ymidleft, xmidright, ymidright, x2, y2);
//				// Add feet
//				path.append(sourceShape, false);
//				path.append(targetShape, false);
                shape = path;
                break;
            case TIE:
                path = new GeneralPath();
                int dx = x2 - x1;
                int dy = y2 - y1;
                float length = (float) Math.sqrt(dx * dx + dy * dy);
                float unitx = dx / length;
                float unity = dy / length;
                path.moveTo(x1 + unity * _scale, y1 - unitx * _scale);
                path.lineTo(x2 + unity * _scale, y2 - unitx * _scale);
                path.moveTo(x1 - unity * _scale, y1 + unitx * _scale);
                path.lineTo(x2 - unity * _scale, y2 + unitx * _scale);
//				// Add feet
//				path.append(sourceShape, false);
//				path.append(targetShape, false);
                shape = path;
        }
        return shape;
    }
    
    public Shape getShape() {
    	return getShape(false);
    }
    
	public void translateTo(float dx, float dy) {
		_source.translateTo(dx, dy);
		_target.translateTo(dx, dy);
	}    

    public java.util.Set getHandles() {
		// Set handles
		java.util.Set _handles = new java.util.LinkedHashSet();
		_handles.add(getSource().getHandle());
		_handles.add(getTarget().getHandle());
    	return _handles;
    }
    public void setType(int type) {
        _type = type;
        if (type == TIE)
        	setThickness(DiagramCanvas.THIN.getLineWidth());
        else
        	setThickness(DiagramCanvas.DEFAULT.getLineWidth());        	
    }
    
    public int getType() {
        return _type;
    }
    
    public ConcreteSpiderFoot getSource() {
        return _source;
    }
    
    public ConcreteSpiderFoot getTarget() {
        return _target;
    }    
    
    public double getLength() {
		double x1 = _source.getX();
		double y1 = _source.getY();
		double x2 = _target.getX();
		double y2 = _target.getY();
		double value = (x2-x1)*(x2-x1)+(y2-y1)*(y2-y1);
		return value;
    }
    
	public DiagramItem getClone() {
		ConcreteSpiderFoot newSource = (ConcreteSpiderFoot)_source.getClone();
		ConcreteSpiderFoot newTarget = (ConcreteSpiderFoot)_target.getClone();
		ConcreteSpiderLeg newLeg = new ConcreteSpiderLeg(_label, newSource, newTarget, _type);
		newLeg.setAttributes(this);
		return newLeg;
	}
	
	/**
	 * @return string representation of this leg
	 */
	public String toString() {
		return "(" + getSource() + "," + getTarget() + ")";
	}
	
	/**
	 * 
	 * @return a user-friendly description of this item
	 */
	/*
	 * test:
	 *   connector is leg
	 *                strand with label
	 *                tie
	 */
	public String id() {
		String s;
		switch(this.getType()) {
			default:
			case NORMAL:
				s = "leg";
				break;
				
			case STRAND:
				s = "strand";
				break;
			
			case TIE:
				s = "tie";
				break;
		}
		if (getLabel().length()!=0)
			s += " " + getLabel();
		return s + " between " + getSource().id() + " and " + getTarget().id();
	}
	


    //
    // Local properties
    //
    private int _type;
    private int _scale = 20;
    private ConcreteSpiderFoot _source;
    private ConcreteSpiderFoot _target;
    
    /*
    // test program
    // ------------

    private static void pr(String s) {
      System.out.println(s);
    }

    private static void hd(String s) {
      System.out.println("\n" + s + "\n");
    }

    private static void pg() {
      try {
        int i;
        System.out.print("next page? (CR)");
        do i = System.in.read(); while (i != '\n');
        System.out.println();
      }
      catch (Exception e) {System.out.println("raised exception : " + e);}
    }

    public static void main(String args[]) throws Exception {
    	hd("Tests of toString, id");
      	ConcreteSpiderFoot csf1 = 
    		new ConcreteSpiderFoot("x",ConcreteSpiderFoot.NORMAL);
      	csf1.setLocation(1,2);
      	pr("csf1="+csf1);
      	ConcreteSpiderFoot csf2 = 
    		new ConcreteSpiderFoot("y",ConcreteSpiderFoot.NORMAL);
      	csf2.setLocation(3,4);
      	pr("csf2="+csf2);
    	ConcreteSpiderLeg csl = 
    		new ConcreteSpiderLeg("t",csf1,csf2,ConcreteSpiderLeg.NORMAL);
    	pr("csl="+csl);
    	pr("csl.id()="+csl.id());
    	pr("csl.hashCode()="+csl.hashCode());
    	pg();
    	
    	{
	    	ConcreteSpiderLeg csl2 = 
	    		new ConcreteSpiderLeg("t",csf1,csf2,ConcreteSpiderLeg.STRAND);
	    	pr("csl2="+csl2);
	    	pr("csl2.id()="+csl2.id());
	    	pr("csl2.hashCode()="+csl2.hashCode());
	    	pg();
	    	
	    	ConcreteSpiderLeg csl3 = 
	    		new ConcreteSpiderLeg("t",csf1,csf2,ConcreteSpiderLeg.TIE);
	    	pr("csl3="+csl3);
	    	pr("csl3.id()="+csl3.id());
	    	pr("csl3.hashCode()="+csl3.hashCode());
	    	pg();
    	}
    	
    	{
    		hd("Tests of getClone, equals");
    		ConcreteSpiderLeg cslc = (ConcreteSpiderLeg) csl.getClone();
    		pr("cslc="+cslc);
    		pr("cslc.hashCode()="+cslc.hashCode());
    		pr("cslc.id()="+cslc.id());
    		pg();
    		
    		pr("csl.equals(cslc)="+csl.equals(cslc));
    		pr("cslc.equals(csl)="+cslc.equals(csl));
    		pg();
    		
    		pr("csl.equals(null)="+csl.equals(null));
    		pr("csl.equals(csf1)="+csl.equals(csf1));
    		pg();
    		
    	}
    	
    }
    */
}
