package rwd.gui.editor.listeners.diagramcontrollers;

import java.awt.Cursor;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;

import javax.swing.JFrame;

import rwd.gui.editor.DiagramCanvas;
import rwd.gui.editor.viewers.DiagramViewerComponent;

/**
 * The controller for user gestures on a constraint diagram.
 * This controller is implemented using the state pattern.
 * It has a current state object which can be plugged in.
 * This controller delegates some user gestures to its current state object,
 * allowing a flexible response based on the current state.
 * State objects can update the current state to a different state.
 * States are represented by classes which implement the DiagramControllerState
 * interface.
 * <P>
 * This class disambiguates mouse gestures before passing them on to its
 * current state object.
 * If the mouse is pressed, dragged* and released on approximately the same pixel
 * (+-MOUSE_LATITUDE pixels), a single MouseClicked event will be passed on.
 * If the mouse is pressed, dragged* and released, and the mouse moves 
 * outside the MOUSE_LATITUDE then a 
 * MousePressed MouseDragged+ MouseReleased 
 * sequence will be passed to the current state object.
 * The disambiguation process makes it easier for the state object to
 * interpret the user's gestures, and avoids unexpected effects.
 * 
 * @author Richard Bosworth
 * 
 * @version 19-Apr-2005 12:09
 */
public class ConstraintDiagramController 
	implements DiagramController
{
	// the diagram canvas being controlled by this controller
	DiagramCanvas myDiagramCanvas;
	// the current state: initially the StartState
	private DiagramControllerState myState;
	
	
	// constructor
	// -----------
	
	/**
	 * Post: new controller has been created for specified diagram 
	 *       with state set to start state
	 * @param cd = diagram canvas for this controller
	 */
	/*
	 * test: once thru
	 */
	public ConstraintDiagramController(DiagramCanvas dc)
	{
		myDiagramCanvas = dc;
		myState = new StartState(this);
	}
	
	// attributes
	// ----------
	
	/**
	 * @return Canvas controlled by this controller
	 */
	/*
	 * test: once thru
	 */
	public DiagramCanvas getDiagramCanvas()
	{
		return myDiagramCanvas;
	}
	
	
	/**
	 * @return current state of this controller
	 */
	/*
	 * test: once thru
	 */
	public DiagramControllerState getState()
	{
		return myState;
	}
	
	/**
	 * @return a string representation of the controller
	 */
	/*
	 * test: once thru
	 */
	public String toString()
	{
		return super.toString() + ": current state=" + myState;
	}
	
	
	// operations
	// ----------
	
	/**
	 * Post: controller's state has been set to s.
	 * @param s required state of controller (must be non-null)
	 */
	/*
	 * test: once thru
	 */
	public void setState(DiagramControllerState s)
	{
		myState = s;
	}
	
	
	// the user gestures
	// -----------------
	
	private static final int GESTURE_STARTED = 0;
	private static final int GESTURE_CONTINUED = 1;
	private static final int GESTURE_ENDED = 2;
	  
	private int mx;  // mouse x position in pixel integral coordinates
	private int my;  // mouse y position in pixel integral coordinates
	private static final int MOUSE_LATITUDE = 1;  // +- pixels allowed for click
	  
	private int mouseState = GESTURE_ENDED;

	/**
	 * @see DiagramController#mouseClicked(java.awt.event.MouseEvent)
	 */
	/*
	 * test: once thru
	 */
	public void mouseClicked(MouseEvent e) 
	{
		// ignore awt mouse-clicks, which are generated if the mouse is released
		// near where it was pressed, irrespective of where it has been 
		// in the meantime
	};
	
	/**
	 * @see DiagramController#mousePressed(java.awt.event.MouseEvent)
	 */
	/*
	 * test: once thru
	 */
	public void mousePressed(MouseEvent e) 
	{
		mouseState = GESTURE_STARTED;
	    mx = e.getX();
	    my = e.getY();
	};
	
	/**
	 * @see DiagramController#mouseReleased(java.awt.event.MouseEvent)
	 */
	/*
	 * test:
	 *   release on same pixel as press
	 *   release on near pixel to press
	 *   release far from press with intervening drags
	 *   release without press
	 */
	public void mouseReleased(MouseEvent e) 
	{
	    switch (mouseState)
	    {
	    	case GESTURE_STARTED:
	    	    if ((Math.abs(e.getX()-mx)<= MOUSE_LATITUDE) &&
	    		        (Math.abs(e.getY()-my)<= MOUSE_LATITUDE))
	    	    {
	    			mouseState = GESTURE_ENDED; 
	    	    	myState.mouseClicked(e);
	    	    }
	    		else
	    		{
	    			mouseState = GESTURE_ENDED;
	    			myState.mouseReleased(e);
	    		}
	    		break;
	      
	    	case GESTURE_CONTINUED:
	    		mouseState = GESTURE_ENDED;
	    		myState.mouseReleased(e);
	    		break;
	        
	    	case GESTURE_ENDED:
	    	default:
	    		break;  // ignore for stability      
	    }
	  }
	
	/**
	 * @see DiagramController#mouseEntered(java.awt.event.MouseEvent)
	 */
	/*
	 * test: once thru
	 */
	public synchronized void mouseEntered(MouseEvent e) {
		e.consume();
		Cursor c = new Cursor(Cursor.CROSSHAIR_CURSOR);
		myDiagramCanvas.setCursor(c);
	}

	
	/**
	 * @see DiagramController#mouseExited(java.awt.event.MouseEvent)
	 */
	/*
	 * test: once thru
	 */
	public synchronized void mouseExited(MouseEvent e) {
		e.consume();
		Cursor def = new Cursor(Cursor.DEFAULT_CURSOR);
		myDiagramCanvas.setCursor(def);
	}


	
	/**
	 * @see DiagramController#mouseDragged(java.awt.event.MouseEvent)
	 */
	/*
	 * test:
	 *   drag without preceding press
	 *   drag with press within latitude
	 *   drag with press outside latitude
	 *   
	 */
	public void mouseDragged(MouseEvent e) 
	{
	          
        switch (mouseState)
        {
        	case GESTURE_STARTED:
        	    if ((Math.abs(e.getX()-mx)> MOUSE_LATITUDE) ||
        		        (Math.abs(e.getY()-my)> MOUSE_LATITUDE))
        	    {
	        		mouseState = GESTURE_CONTINUED;
	        		myState.mousePressed(e);
	        		myState.mouseDragged(e);
        	    }
        		break;
          
        	case GESTURE_CONTINUED:
        		myState.mouseDragged(e);
        		break;
            
        	case GESTURE_ENDED:
        	default:
        		break;  // ignore for stability
        }
	}
	
	/**
	 * @see DiagramController#mouseMoved(java.awt.event.MouseEvent)
	 */
	/*
	 * test: once thru
	 */
	public synchronized void mouseMoved(MouseEvent e) {
		e.consume();
	}


	
	/**
	 * @see DiagramController#keyTyped(java.awt.event.KeyEvent)
	 */
	/*
	 * test: once thru
	 */
	public void keyTyped(KeyEvent e) 
	{
		myState.keyTyped(e);
	};
	
	/**
	 * @see DiagramController#keyPressed(java.awt.event.KeyEvent)
	 */
	/*
	 * test: once thru
	 */
	public void keyPressed(KeyEvent e) {};
	
	/**
	 * @see DiagramController#keyReleased(java.awt.event.KeyEvent)
	 */
	/*
	 * test: once thru
	 */
	public void keyReleased(KeyEvent e) {};
	
	
	// test program
    // ------------

    private static void pr(String s) {
      System.out.println(s);
    }

    private static void hd(String s) {
      System.out.println("\n" + s + "\n");
    }

    private static void pg() {
      try {
        int i;
        System.out.print("next page? (CR)");
        do i = System.in.read(); while (i != '\n');
        System.out.println();
      }
      catch (Exception e) {System.out.println("raised exception : " + e);}
    }

	public class TestFrame
		extends JFrame
	{
		public TestFrame()
		{
			setTitle("Test frame");
			
	    	DiagramViewerComponent dvc = new DiagramViewerComponent(100,200,true);
			DiagramCanvas myCanvas = new DiagramCanvas(dvc);
			ConstraintDiagramController cdc = 
				new ConstraintDiagramController(myCanvas);
			myCanvas.addKeyListener(cdc);
			myCanvas.addMouseListener(cdc);
			myCanvas.addMouseMotionListener(cdc);
			
			this.getContentPane().add(myCanvas);
			this.show();
		}
	}
    public static void main(String args[]) throws Exception {
    	hd("Test of constructor");
    	DiagramViewerComponent dvc = new DiagramViewerComponent(100,200,true);
    	DiagramCanvas dc = new DiagramCanvas(dvc);
    	ConstraintDiagramController cdc = new ConstraintDiagramController(dc);
    	pr("cdc="+cdc);
    	pr("cdc.getDiagramCanvas()="+cdc.getDiagramCanvas());
    	pr("cdc.getState()="+cdc.getState());
    	pg();
    	
    	
    	{
	    	hd("Test of setState");
	    	DiagramControllerState newState = new StartState(cdc) {};
	    	pr("set cdc to new state");
	    	cdc.setState(newState);
	    	pr("cdc="+cdc);
	    	pg();
	    	
	    	cdc.setState(new StartState(cdc));
	    	pr("cdc="+cdc);
	    	pg();
    	}
    	
    	{
	    	hd("Test of mouse gesture disambiguating");
	    	TestFrame tf = cdc.new TestFrame();
	    	pr("tf="+tf);
	    	pg();
	    	
    	}
    	
    }
    
}
