package rwd.metrics;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.Vector;

import rwd.concretesyntax.ConcreteUnitaryDiagram;
import rwd.concretesyntax.constraintdiagram.ConcreteArc;
import rwd.concretesyntax.constraintdiagram.ConcreteArcEnd;
import rwd.concretesyntax.spiderdiagram.ConcreteSpider;
import rwd.concretesyntax.spiderdiagram.ConcreteSpiderFoot;
import rwd.concretesyntax.spiderdiagram.ConcreteSpiderLeg;
import rwd.deltas.AngularResolutionDelta;
import rwd.deltas.Delta;

/**
 * @author Octavian Patrascoiu
 * 
 * Angular Resolution. We want to penalize two edges emanating from the same node 
 * that are too close together.
 * 
 * Inverse of summed square of distance
 */
public class AngularResolution
	extends SpiderMetric 
{
	public AngularResolution(ConcreteUnitaryDiagram cud, double weighting) {
		super(cud, weighting);
		name = "AngularResolutionMetric";
		key = "ARC";
	}
	public AngularResolution(ConcreteUnitaryDiagram cud) {
		super(cud);
		name = "AngularResolutionMetric";
		key = "ARC";
	}

	public void initDeltas(ConcreteUnitaryDiagram cud) {
		if (cud == null) return;
		// For each spider
		deltas = new Vector();
		Set spiders = cud.getConcreteSpiders();
		Iterator i = spiders.iterator();
		while (i.hasNext()) {
			ConcreteSpider spider = (ConcreteSpider)i.next();
			// For each pair of distinct legs
			List legs = new Vector(spider.getLegs());
			for(int j=0; j<legs.size(); j++) {
				ConcreteSpiderLeg leg1 = (ConcreteSpiderLeg)legs.get(j);
				for(int k=j+1; k<legs.size(); k++) {
					ConcreteSpiderLeg leg2 = (ConcreteSpiderLeg)legs.get(k);
					// If they share a foot
					ConcreteSpiderFoot s1 = leg1.getSource();
					ConcreteSpiderFoot t1 = leg1.getTarget();
					ConcreteSpiderFoot s2 = leg2.getSource();
					ConcreteSpiderFoot t2 = leg2.getTarget();
					if (s1 == s2) { 
						Delta delta = new AngularResolutionDelta(s1, t1, t2);
						deltas.add(delta);
					} else if (s1 == t2) {
						Delta delta = new AngularResolutionDelta(s1, t1, s2);
						deltas.add(delta);
					} else if (t1 == s2) {
						Delta delta = new AngularResolutionDelta(t1, s1, t2);
						deltas.add(delta);
					} if (t1 == t2) {
						Delta delta = new AngularResolutionDelta(t1, s1, s2);
						deltas.add(delta);
					}
				}
			}
		}
		// For each pair (leg, arc) that are close enough
		List arcs = new Vector(cud.getConcreteArcs());
		List legs = new Vector(cud.getConcreteLegs());
		for(int i1=0; i1<arcs.size(); i1++) {
			ConcreteArc arc = (ConcreteArc)arcs.get(i1);
			ConcreteArcEnd s1 = arc.getStart();
			ConcreteArcEnd t1 = arc.getEnd();
			for(int i2=0; i2<legs.size(); i2++) {
				ConcreteSpiderLeg leg = (ConcreteSpiderLeg)legs.get(i2);
				ConcreteSpiderFoot s2 = leg.getSource();
				ConcreteSpiderFoot t2 = leg.getTarget();
				if (near(s1, s2)) {
					Delta delta = new AngularResolutionDelta(s1, t1, t2);
					deltas.add(delta);				
				} else if (near(s1, t2)) {
					Delta delta = new AngularResolutionDelta(s1, t1, s2);
					deltas.add(delta);				
				} else if (near(t1, s2)) {
					Delta delta = new AngularResolutionDelta(t1, s1, t2);
					deltas.add(delta);				
				} else if (near(t1, t2)) { 
					Delta delta = new AngularResolutionDelta(t1, s1, s2);
					deltas.add(delta);				
				}
			}
		}
	}

	protected boolean near(ConcreteArcEnd item1, ConcreteSpiderFoot item2) {
		double x1 = item1.getX();
		double y1 = item1.getY();
		double x2 = item2.getX();
		double y2 = item2.getY();
		double d = (x2-x1)*(x2-x1)+(y2-y1)*(y2-y1);
		if (Math.abs(d) < 10)
			return true;
		return false;	
	}
}
