package rwd.ocl.constraintdiagram;

import java.util.*;

import rwd.concretesyntax.*;
import rwd.concretesyntax.eulerdiagram.*;
import rwd.concretesyntax.spiderdiagram.*;
import rwd.concretesyntax.constraintdiagram.*;

import rwd.fopl.FoplVisitor;
import rwd.fopl.ast.*;
import rwd.fopl.visitors.OclVisitor;
import rwd.fopl.visitors.StringVisitor;

import uk.ac.kent.cs.kmf.util.ILog;
//import uk.ac.kent.cs.kmf.util.OutputStreamLog;

/**
 * The abstract representation of a reading tree.
 * @author Octavian Patrascoiu, rnb
 * @version 8.9.05  14:13
 *
 */
public class ReadingTree {
	//
	// Constructors
	//
	public ReadingTree(String exp) {
		this.log = null;
		root = translate(exp);
	}
	public ReadingTree(String exp, ILog log) {
		this.log = log;
		root = translate(exp);
	}
	
	
	/**
	 * 
	 * @return names of nodes excluding the root (PTC)
	 */
	/*
	 * test:
	 *   tree of no nodes
	 *   tree of two nodes
	 */
	public List getNodeNames() {
		List result = new ArrayList();
		Iterator nit = root.getDescendants().iterator();
		while (nit.hasNext()) {
			result.add(((ReadingTreeNode)nit.next()).getInfo());
		}
		return result;
	}

	
	/*
	syntax of reading tree string:
	  start ->  forest 
	  forest -> tree*
	  tree -> '(' ID forest ')'
	*/
	
	/*
	 * pre:  exp = reading tree string expression
	 * post: returns reading tree corresponding to exp, or has given message on log
	 *         if syntax of exp is incorrect
	 */
	/*
	 * test: once thru with tree of several nodes
	 */
	private ReadingTreeNode translate(String exp) {
		lexer = new StringLexer(exp);
		tok = lexer.nextToken();
		ReadingTreeNode myRoot = new ReadingTreeNode("PTC");
		computeForest(myRoot);
		if (tok.code != Token.END){
			if (log != null)
				log.reportError("Syntax error in reading tree");
		}
		return myRoot;
	}
	
	/*
	 * pre: n = current node of this tree
	 *      lexer = lexer for string of this tree
	 *      tok = next token of string
	 * post: if translation OK, descendants have been added to current node,
	 *         else message has been sent to log file (if it exists).
	 */
	/*
	 * test: no sons
	 *       two sons
	 *       syntax error
	 */
	protected void computeForest(ReadingTreeNode n) {
		while (tok.isLP()) {
			ReadingTreeNode tree = computeTree();
			if (tree != null) 
				n.addSon(tree);
		}
	}
	
	/*
	 * pre: lexer = lexer for string of this tree
	 *      tok = next token of string
	 * post: if translation OK, returns new reading tree
	 *       else message has been sent to log file (if it exists),
	 *            and returns null.
	 */
	/*
	 * test:
	 *   )
	 *   (())
	 *   ()
	 *   (fred)
	 *   (fred!
	 *   (fred(jim)(joe))
	 * 
	 */
	protected ReadingTreeNode computeTree() {
		ReadingTreeNode myRoot = null;
		// (
		if (tok.isLP()) {
			// Skip (
			tok = lexer.nextToken();
			// Get name and skip
			if (tok.isID()) {
				String name = tok.info;
				tok = lexer.nextToken();
				myRoot = new ReadingTreeNode(name);
			}
			else {
				if (log != null) {
					log.reportError("Missing node ID");
				}
			}
			// Get sons if possible
			if (myRoot != null)
				computeForest(myRoot);
			
			// )
			if (tok.isRP()) {
				// Skip )
				tok = lexer.nextToken();
			} else {
				if (log != null)
					log.reportError("Missing ')'");				
			}
		// Error
		} else {
			if (log != null)
				log.reportError("Syntax error in reading tree");
		}
		return myRoot;
	}
	
	/**
	 * 
	 * @param label1 label of possible ancestor
	 * @param label2 label of possible descendant
	 * @return true iff label1 is the label of an ancestor of a node with label2
	 *           in this reading tree
	 */
	/*
	 * test:
	 *   label2 not in tree
	 *   label2 in tree
	 *     label1 is not in tree
	 *     label1 is in tree
	 *     label1 is in tree, but is not an ancestor
	 */
	public boolean isAncestor(String label1, String label2) {
		ReadingTreeNode son = find(label2,root);
		if (son == null) return false;
		return son.hasAncestor(label1);
	}
	
	/**
	 * 
	 * @param label label to be found
	 * @param n root of tree to be searched (may be null)
	 * @return node of tree with given label, or null if none
	 */
	/*
	 * n is null
	 *      single node with right label
	 *      single node with wrong label
	 *      node with grandson of second son with right label
	 *      node with grandson of second son with wrong label
	 */
	public ReadingTreeNode find(String label, ReadingTreeNode n) {
		if (n == null) return null;
		if (n.getInfo().equals(label)) return n;
		Iterator sit = n.getSons().iterator();
		while (sit.hasNext()) {
			ReadingTreeNode son = (ReadingTreeNode) sit.next();
			ReadingTreeNode result = find(label,son);
			if (result != null) return result;
		}
		return null;
	}
	
	/**
	 * 
	 * @param label label to be found in this tree
	 * @return node of tree with given label, or null if none
	 */
	/*
	 * test: once thru
	 */
	public ReadingTreeNode find(String label) {
		return find(label,root);
	}
	
	
	//
	// FOPL and OCL
	//
	protected String readFOPL(ConcreteUnitaryDiagram cud) {
		if (cud == null)
			return "Cannot read diagram 'null'";
		Expression ast = readAST(cud);
		FoplVisitor visit = new StringVisitor();
		String res = (String)ast.accept(visit, null);
		return res;
	}
	protected String readOCL(ConcreteUnitaryDiagram cud) {
		if (cud == null)
			return "Cannot read diagram 'null'";
		// Compute expression
		Expression ast = readAST(cud);
		FoplVisitor visit = new OclVisitor();
		String res = (String)ast.accept(visit, null);
		return res;
	}
	protected String getPTCinFOPL(ConcreteUnitaryDiagram cud) {
		if (cud == null)
			return "Cannot compute PTC for diagram 'null'";
		// Compute PTC
		Expression ast = makePTC(cud);
		FoplVisitor visit = new StringVisitor();
		String res = (String)ast.accept(visit, null);
		return res;
	}
	protected String getPTCinOCL(ConcreteUnitaryDiagram cud) {
		if (cud == null)
			return "Cannot compute PTC for diagram 'null'";
		// Compute PTC
		Expression ast = makePTC(cud);
		FoplVisitor visit = new OclVisitor();
		String res = (String)ast.accept(visit, null);
		return res;
	}
	
	//
	// AST FOPL section
	//
	/*
	 * test: 
	 *   null cud
	 *   tree with root only
	 *   tree with several sons
	 */
	public Expression readAST(ConcreteUnitaryDiagram cud) {
		if (cud == null)
			return null;
		if (root.getSons().isEmpty()) return null;
		// Compute a conjunction
		List exps = new Vector();
		// Add PTC
		Expression ptc = makePTC(cud);
		if (ptc != null)
			exps.add(ptc);
		// Init common conditions
		initCondsAST(cud);
		// Read each tree
		Iterator sit = root.getSons().iterator();
		while (sit.hasNext()) {
			ReadingTreeNode root = (ReadingTreeNode) sit.next();
			Expression exp = makeIndependentExpression(root, cud);
			exps.add(exp);
		}
		// Make the conjunction
		return makeExpression(exps, AND);
	}
	public Expression makePTC(ConcreteUnitaryDiagram cud) {
		// Init items
		initItems(cud);
		// Compute PTC as conjunction
		List conds = new Vector();
		// Make a condition for each missing zone
		Set allZones = cud.makeAllConcreteZones();
		Set zones = cud.getConcreteZones();
		Iterator i = allZones.iterator();
		while (i.hasNext()) {
			ConcreteZone zone = (ConcreteZone)i.next();
			boolean found = search(zone, zones);
			if (!found) {
				// cond = text(zone, cd) EQ EMPTY
				Expression left = textAST(zone, cud, items); 
				Constant right = new ConstantImpl();
				right.setValue(EMPTY);
				BinaryExpression eq = new BinaryExpressionImpl();
				eq.setOperator(EQ);
				eq.setLeft(left);
				eq.setRight(right);
				conds.add(eq);
			}
		}
		return makeExpression(conds, AND);
	}
	protected Expression makeIndependentExpression(ReadingTreeNode root, ConcreteUnitaryDiagram cud) {
		// Make the initial condition
		Expression init = makeExpression(initConds, AND);
		// Read the diagram
		Expression read = readAST(root, cud);
		// Assemble them
		if (init != null) {
			if (read != null) {
				BinaryExpression exp = new BinaryExpressionImpl();
				exp.setOperator(AND);
				exp.setLeft(init);
				exp.setRight(read);
				return exp; 
			} else {
				return init;
			}
		} else {
			if (read != null) {
				return read; 
			} else {
				Constant exp = new ConstantImpl();
				exp.setValue("true");
				return exp;
			}
		}
	}
	protected void initCondsAST(ConcreteUnitaryDiagram cud) {
		// Add all arrows which have both source and target contours
		Iterator i = cud.getConcreteArcs().iterator();
		while (i.hasNext()) {
			ConcreteArc arc = (ConcreteArc)i.next();
			DiagramItem source = arc.getSource();
			DiagramItem target = arc.getTarget();
			if (source instanceof ConcreteContour && target instanceof ConcreteContour) {
				items.add(arc);
				if (isGivenContour(source) && isGivenContour(target)) {
					// cond = textAST(source, cud)+POINT+textAST(arc, cud)+EQ+textAST(target, cud);
					BinaryExpression dot = new BinaryExpressionImpl();
					dot.setOperator(DOT);
					dot.setLeft(textAST(source, cud, items));
					dot.setRight(textAST(arc, cud, items)); 
					BinaryExpression eq = new BinaryExpressionImpl();
					eq.setOperator(EQ);
					eq.setLeft(dot);
					eq.setRight(textAST(target, cud, items));
					initConds.add(eq);
				}
			}
		}		
	}
	protected Expression readAST(ReadingTreeNode root, ConcreteUnitaryDiagram cud) {
		if (root == null)
			return null;
		else {
			// Compute spider
			String spName = root.info;
			ConcreteSpider sp = cud.getConcreteSpider(spName);
			// Make quant
			Cuantifier quant = makeQuantAST(root, sp, cud);
			// Make condition
			Expression cond = makeCondAST(root, sp, cud);
			// Return result
			quant.setCond(cond);
			return quant;
		}
	}
	protected Cuantifier makeQuantAST(ReadingTreeNode root, ConcreteSpider sp, ConcreteUnitaryDiagram cud) {
		// Add feet
		Set feet = sp.getFeet();
		items.addAll(feet);
		// Compute quantifier
		Cuantifier quant = new CuantifierImpl();
		if (sp.isUniversal()) {
			quant.setKind(EVERY);
		} else if (sp.isExistential()) {
			quant.setKind(EXISTS);
		} else {
			quant.setKind(SCHROD);
		}
		quant.setVar(root.info);
		// Conpute region
		Set zones = sp.getHabitat(cud);
		Expression reg = textAST(zones, cud, items);
		// Compute result
		quant.setDom(reg);
		// Add new items derived from spider
		newItems = new Vector(getNewItems(cud));
		items.addAll(newItems);
		return quant;
	}
	protected Expression makeCondAST(ReadingTreeNode root, ConcreteSpider sp, ConcreteUnitaryDiagram cud) {
		// Compute condition
		List conds = new Vector();
		// For each new arrow
		for(int i=0; i<newItems.size(); i++) {
			Object obj = newItems.get(i);
			if (obj instanceof ConcreteArc) {
				ConcreteArc arc = (ConcreteArc)obj;
				DiagramItem source = arc.getSource();
				DiagramItem target = arc.getTarget();
				// 1 - page 20
				if (source instanceof ConcreteSpiderFoot && target instanceof ConcreteSpiderFoot) {
//					String cond = text(source, cud)+POINT+text(arc, cud)+EQ+text(target, cud);
					BinaryExpression dot = new BinaryExpressionImpl();
					dot.setOperator(DOT);
					dot.setLeft(textAST(source, cud, items));
					dot.setRight(textAST(arc, cud, items));					
					BinaryExpression eq = new BinaryExpressionImpl();
					eq.setOperator(EQ);
					eq.setLeft(dot);
					eq.setRight(textAST(target, cud, items));					
					conds.add(eq);
				// 2 or 3 page 20, 21
				} else if (isDerivedContour(target)) {
					// 2
					Set spiders = spidersWithHabitatChanged((ConcreteContour)target, cud);
					if (spiders.size() == 0) {
						// left = text(DependenceGraph.rsmall((ConcreteContour)target, cud), cud);
						// middle = text(source, cud)+POINT+text(arc, cud);
						// right = text(DependenceGraph.rbig((ConcreteContour)target, cud), cud);
						// cond = left+INCLUDED_OR_EQUAL+middle+INCLUDED_OR_EQUAL+right;
						Expression left = textAST(DependenceGraph.rsmall((ConcreteContour)target, cud), cud, items);
						Expression right = textAST(DependenceGraph.rbig((ConcreteContour)target, cud), cud, items);
						BinaryExpression middle = new BinaryExpressionImpl();
						middle.setOperator(DOT);
						middle.setLeft(textAST(source, cud, items));
						middle.setRight(textAST(arc, cud, items));
						BinaryExpression cond1 = new BinaryExpressionImpl();
						cond1.setOperator(INCLUDED_OR_EQUAL);
						cond1.setLeft(left);
						cond1.setRight(middle);
						BinaryExpression cond2 = new BinaryExpressionImpl();
						cond2.setOperator(INCLUDED_OR_EQUAL);
						cond2.setLeft(middle);
						cond2.setRight(right);
						BinaryExpression cond = new BinaryExpressionImpl();
						cond.setOperator(AND);
						cond.setLeft(cond1);
						cond.setRight(cond2);
						conds.add(cond);
					// 3
					} else {
						Iterator j=spiders.iterator();
						while (j.hasNext()) {
							ConcreteSpider sp1 = (ConcreteSpider)j.next();
							// cond = text(sp1, cud)+BELONGS+text(sp1.getHabitat(cud), cud);
							BinaryExpression cond = new BinaryExpressionImpl();
							cond.setOperator(BELONGS);
							cond.setLeft(textAST(sp1, cud, items));
							cond.setRight(textAST(sp1.getHabitat(cud), cud, items));
							conds.add(cond);
						}
					}
				} else {
				}
			}
		}
		// 4 page 21
		Set habitat = sp.getHabitat(cud);
		Iterator j = items.iterator();
		while(j.hasNext()) {
			Object obj1 = j.next();
			if (obj1 instanceof ConcreteSpider) {
				ConcreteSpider sp1 = (ConcreteSpider)obj1;
				Set habitat1 = sp1.getHabitat(cud);
				habitat1.retainAll(habitat);
				if (habitat1.size() != 0) {
					// sp1 touched by sp
					Iterator k = habitat1.iterator();
					while(k.hasNext()) {
						ConcreteZone cz = (ConcreteZone)k.next();
						// cond = text(sp, cud)+NE+text(sp1, cud);
						BinaryExpression cond = new BinaryExpressionImpl();
						cond.setOperator(NE);
						cond.setLeft(textAST(sp, cud, items));
						cond.setRight(textAST(sp1, cud, items));
						conds.add(cond);
					}
				}
			}
		}
		// 5 page 21
		Set shadedZones = DependenceGraph.potentialShadedZonesConsequences(sp, cud);
		j = shadedZones.iterator();
		while(j.hasNext()) {
			Object obj1 = j.next();
			if (obj1 instanceof ConcreteZone) {
				ConcreteZone cz = (ConcreteZone)obj1;
				Iterator k = items.iterator();
				while (k.hasNext()) {
					Object obj2 = k.next();
					if (obj2 instanceof ConcreteSpider) {
						ConcreteSpider sp1 = (ConcreteSpider)obj2;
						Set habitat1 = sp1.getHabitat(cud);
						if (habitat1.contains(cz)) {
							// cond = text(cz, cud)+MINUS+"{"+text(sp1, cud)+"}="+EMPTY;
							BinaryExpression min = new BinaryExpressionImpl();
							min.setOperator(MINUS);
							min.setLeft(textAST(cz, cud, items));
							min.setRight(textAST(sp1, cud, items));
							Constant con = new ConstantImpl();
							con.setValue(EMPTY);
							BinaryExpression cond = new BinaryExpressionImpl();
							cond.setOperator(EQ);
							cond.setLeft(min);
							cond.setRight(con);
							conds.add(cond);
						}
					}
				}
			}
		}
		 
		// Add spider
		items.add(sp);
		// Add reading for each son
		List sons = root.sons; 
		for(int i=0; i<sons.size(); i++) {
			ReadingTreeNode son = (ReadingTreeNode)sons.get(i);
			Expression reading = readAST(son, cud);
			conds.add(reading);
		}
		Expression cond = null;
		if (conds.size() == 0) {
			cond = new ConstantImpl();
			((Constant)cond).setValue("true");
		} else {
			cond = makeExpression(conds, AND);
		}
		return cond;
	}
	
	/** Compute text function */
	protected Expression textAST(Object obj, ConcreteUnitaryDiagram cud, Set items) {
		// Concrete elements
		if (obj == null) return null;
		if (obj instanceof ConcreteContour) {
			ConcreteContour c = (ConcreteContour)obj;
			if (!c.isDerived()) {
				Type res = new TypeImpl();
				String type = c.getLabel(); 
				res.setValue(type);
				return res; 
			} else {
				// Find arc with c as target
				Iterator i = items.iterator();
				while (i.hasNext()) {
					Object obj1 = i.next();
					if (obj1 instanceof ConcreteArc) {
						ConcreteArc arc = (ConcreteArc)obj1;
						DiagramItem source = arc.getSource();
						DiagramItem target = arc.getTarget();
						if (target == c) {
							// Foound
							if (isGivenContour(source)) {
								// text(source, cud)+POINT+text(arc, cud);
								BinaryExpression res = new BinaryExpressionImpl();
								res.setOperator(DOT);
								res.setLeft(textAST(source, cud, items));
								res.setRight(textAST(arc, cud, items));
								return res;
							} else if (source instanceof ConcreteSpiderFoot) {
								// text(cud.getConcreteSpider((ConcreteSpiderFoot)source), cud)+POINT+text(arc, cud);
								BinaryExpression res = new BinaryExpressionImpl();
								res.setOperator(DOT);
								res.setLeft(textAST(cud.getConcreteSpider((ConcreteSpiderFoot)source), cud, items));
								res.setRight(textAST(arc, cud, items));
								return res;
							}
						}
					}
				}
				Constant res = new ConstantImpl();
				res.setValue("");
				return res; 
			}
		} else if (obj instanceof ConcreteZone) {
			ConcreteZone cz = (ConcreteZone)obj;
			Set in = new LinkedHashSet();
			Set out = new LinkedHashSet();
			List inCon = cz.getContainingPolygonals();
			for(int j=0; j<inCon.size(); j++) {
				Object obj1 = inCon.get(j);
				if (obj1 instanceof ConcreteContour && items.contains(obj1)) {
					ConcreteContour c = (ConcreteContour)obj1;
					Expression text = textAST(c, cud, items);
					if (text != null)
						in.add(text); 
				}
			}
			List outCon = cz.getExcludingPolygonals();
			for(int j=0; j<outCon.size(); j++) {
				Object obj1 = outCon.get(j);
				if (obj1 instanceof ConcreteContour && items.contains(obj1)) {
					ConcreteContour c = (ConcreteContour)obj1;
					Expression text = textAST(c, cud, items);
					if (text != null)
						out.add(text);
				} 
			}			
			Expression inText = makeExpression(in, INTERSECTION);
			Expression outText = null;
			List outL = new Vector(out);
			for(int i=0; i<outL.size(); i++) {
				Expression text = (Expression)outL.get(i);
				UnaryExpression not = new UnaryExpressionImpl();
				not.setOperator(NOT);
				not.setLeft(text);
				if (i != 0) {
					BinaryExpression exp = new BinaryExpressionImpl();
					exp.setOperator(INTERSECTION);
					exp.setLeft(outText);
					exp.setRight(not);
					outText = exp;
				} else {
					outText = not;
				}
			}
			Expression text = inText;
			if (text != null) {
				if (outText != null) {
					BinaryExpression exp = new BinaryExpressionImpl();
					exp.setOperator(INTERSECTION);
					exp.setLeft(inText);
					exp.setRight(outText);
					text = exp;
				}
			}
			else {
				text = outText;
			}
			// Univers
			if (text == null) {
				Constant con = new ConstantImpl();
				con.setValue(UNIV);
				text = con;
			}
			return text;
		} else if (obj instanceof ConcreteSpiderFoot) {
			ConcreteSpider sp = cud.getConcreteSpider((ConcreteSpiderFoot)obj);
			return textAST(sp, cud, items);
		} else if (obj instanceof ConcreteSpider) {
			Constant con = new ConstantImpl();
			con.setValue(((ConcreteSpider)obj).getLabel());
			return con;
		} else if (obj instanceof ConcreteArc) {
			Constant con = new ConstantImpl();
			con.setValue(((ConcreteArc)obj).getLabel());
			return con;
		// Region
		} else if (obj instanceof Collection) {
			Collection region = (Collection)obj;
			if (region == null || region.size() == 0) {
				Constant con = new ConstantImpl();
				con.setValue(EMPTY);
				return con;
			}
			Set text = new LinkedHashSet();
			Iterator i = region.iterator();
			while (i.hasNext()) {
				Object obj1 = i.next();
				Expression s = textAST(obj1, cud, items);
				text.add(s);				
			}
			Expression res = makeExpression(text, UNION);
			if (res == null) {
				Constant con = new ConstantImpl();
				con.setValue(EMPTY);
				return con;
			}
			return res;
		}
		Constant con = new ConstantImpl();
		con.setValue("");
		return con;
	}
	protected Expression makeExpression(Collection col, String operator) {
		List l = new Vector(col);
		Expression exp = null;
		for(int i=0; i<l.size(); i++) {
			Expression e = (Expression)l.get(i);
			if (i != 0) {
				BinaryExpression temp = new BinaryExpressionImpl();
				temp.setOperator(operator);
				temp.setLeft(exp);
				temp.setRight(e);
				exp = temp;
			} else {
				exp = e;
			}
		}
		return exp;
	}

	//
	// Common section
	//
	protected boolean search(ConcreteZone z, Collection zones) {
		Iterator j = zones.iterator();
		while (j.hasNext()) {
			ConcreteZone z1 = (ConcreteZone)j.next();
			if (z1.equals(z))
				return true;
		}
		return false;
	}
	protected boolean isGivenContour(Object obj) {
		if (obj == null) return false;
		return obj instanceof ConcreteContour && !((ConcreteContour)obj).isDerived();
	}
	protected boolean isDerivedContour(Object obj) {
		if (obj == null) return false;
		return obj instanceof ConcreteContour && ((ConcreteContour)obj).isDerived();
	}
	protected void initItems(ConcreteUnitaryDiagram cud) {
		// Initialize items and cond
		items = new LinkedHashSet();
		initConds = new Vector();
		// Add all contours
		Iterator i = cud.getConcreteContours().iterator();
		while (i.hasNext()) {
			Object obj = i.next();
			if (isGivenContour(obj))
				items.add(obj);
		}
	}
	public Set getNewItems(ConcreteUnitaryDiagram cud) {
		// Initialize result and feet set
		Set res = new LinkedHashSet();
		// Add arrows with source and target in items
		List arrows = new Vector(cud.getConcreteArcs());
		for(int j=0; j<arrows.size(); j++) {
			ConcreteArc arc = (ConcreteArc)arrows.get(j);
			DiagramItem source = arc.getSource();
			DiagramItem target = arc.getTarget();
			if (items.contains(source) && items.contains(target)) {
				res.add(arc);
			}
			if (items.contains(source) && isDerivedContour(target)) {
				res.add(arc);
				res.add(target);
			}
		}
		res.removeAll(items);
		return res;
	}
	protected Set spidersWithHabitatChanged(ConcreteContour c, ConcreteUnitaryDiagram cud) {
		// Scan each spider from items
		Set spiders = new LinkedHashSet();
		Iterator i = items.iterator();
		while (i.hasNext()) {
			Object obj = i.next();
			if (obj instanceof ConcreteSpider) {
				ConcreteSpider sp = (ConcreteSpider)obj;
				Set habitat = sp.getHabitat(cud);
				Set contours = new LinkedHashSet();
				Iterator j = habitat.iterator();
				while (j.hasNext()) {
					ConcreteZone cz = (ConcreteZone)j.next();
					contours.addAll(cz.getContainingPolygonals());
					contours.addAll(cz.getExcludingPolygonals());
				}
				if (contours.contains(c)) {
					spiders.add(sp);
				}
			}
		}
		return spiders;
	}
	/** Compute text function */
	protected String text(Object obj, ConcreteUnitaryDiagram cud) {
		// Concrete elements
		if (obj == null) return "null";
		if (obj instanceof ConcreteContour) {
			ConcreteContour c = (ConcreteContour)obj;
			if (!c.isDerived()) {
				return c.getLabel(); 
			} else {
				// Find arc with c as target
				Iterator i = items.iterator();
				while (i.hasNext()) {
					Object obj1 = i.next();
					if (obj1 instanceof ConcreteArc) {
						ConcreteArc arc = (ConcreteArc)obj1;
						DiagramItem source = arc.getSource();
						DiagramItem target = arc.getTarget();
						if (target == c) {
							// Foound
							if (isGivenContour(source)) {
								return text(source, cud)+DOT+text(arc, cud);
							} else if (source instanceof ConcreteSpiderFoot) {
								return text(cud.getConcreteSpider((ConcreteSpiderFoot)source), cud)+DOT+text(arc, cud);
							}
						}
					}
				}
				return "";
			}
		} else if (obj instanceof ConcreteZone) {
			ConcreteZone cz = (ConcreteZone)obj;
			Set in = new LinkedHashSet();
			Set out = new LinkedHashSet();
			List inCon = cz.getContainingPolygonals();
			for(int j=0; j<inCon.size(); j++) {
				Object obj1 = inCon.get(j);
				if (obj1 instanceof ConcreteContour) {
					ConcreteContour c = (ConcreteContour)obj1;
					String text = text(c, cud);
					if (text.length() != 0)
						in.add(text); 
				}
			}
			List outCon = cz.getExcludingPolygonals();
			for(int j=0; j<outCon.size(); j++) {
				Object obj1 = outCon.get(j);
				if (obj1 instanceof ConcreteContour) {
					ConcreteContour c = (ConcreteContour)obj1;
					String text = text(c, cud);
					if (text.length() != 0)
						out.add(text);
				} 
			}			
			String inText = "";
			Iterator i = in.iterator();
			while (i.hasNext()) {
				String text = (String)i.next();
				inText += text;
				if (i.hasNext()) {
					inText += INTERSECTION;
				}
			}
			String outText = "";
			i = out.iterator();
			while (i.hasNext()) {
				String text = (String)i.next();
				outText += NOT+text;
				if (i.hasNext()) {
					outText += INTERSECTION;
				}
			}
			String text = inText;
			if (text.length() != 0) {
				if (outText.length() != 0)
					text += INTERSECTION+outText;
			}
			else {
				text = outText;
			}
			return text;
		} else if (obj instanceof ConcreteSpiderFoot) {
			ConcreteSpider sp = cud.getConcreteSpider((ConcreteSpiderFoot)obj);
			return text(sp, cud);
		} else if (obj instanceof ConcreteSpider) {
			return ((ConcreteSpider)obj).getLabel();
		} else if (obj instanceof ConcreteArc) {
			return ((ConcreteArc)obj).getLabel();
		// Region
		} else if (obj instanceof Collection) {
			Collection region = (Collection)obj;
			if (region == null || region.size() == 0)
				return EMPTY;
			Set text = new LinkedHashSet();
			Iterator i = region.iterator();
			while (i.hasNext()) {
				Object obj1 = i.next();
				String s = text(obj1, cud);
				text.add(s);				
			}
			i = text.iterator();
			String res = "";
			while (i.hasNext()) {
				String t = (String)i.next();
				res += t;
				if (i.hasNext()) {
					res += UNION;
				}
			}
			if (res.length() == 0)
				return EMPTY;
			return res;
		}
		return "";
	}

	/**
	 * @return string representation of this reading tree.
	 * (Note: the root (Plane Tiling Condition) node is omitted from the string) 
	 */
	/*
	 * test: once thru with tree with several nodes
	 */
	public String toString() {
		String result = "";
		Iterator sit = root.getSons().iterator();
		while (sit.hasNext()) {
			result += ((ReadingTreeNode)sit.next()).toString();
		}
		return result;
	}
	
	/**
	 * @return string representation of this reading tree.
	 */
	/*
	 * test: once thru with tree with several nodes
	 */
	public String toDebugString() {
		return root.toString();
	}
	
	
	//
	// Local properties
	//
	protected ILog log;
	//protected List forest = new Vector();
	protected ReadingTreeNode root;
	protected Lexer lexer;
	protected Token tok;
	protected Set items = new LinkedHashSet();
	protected List newItems = new Vector();
	protected List initConds = new Vector();
	protected List initVars = new Vector();
	
	public static final String EXISTS = "\u2203";
	public static final String EVERY = "\u2200";
	public static final String SCHROD = "\u2298";

	public static final String EQ = "=";
	public static final String NE = "\u2260";

	public static final String DOT = ".";

	public static final String BELONGS = "\u2208";
	public static final String INCLUDED_OR_EQUAL = "\u2286";
	public static final String INCLUDED = "\u2282";

	public static final String NOT = "!";
	public static final String UNION = "\u22c3";
	public static final String INTERSECTION = "\u22c2";
	public static final String MINUS = "-";

	public static final String AND = "\u22c0";
	public static final String OR = "\u22c1";

	public static String UNIV = "_UNIV_";
	public static String EMPTY = "\u2205";
	public static String TRUE = "true";
	public static String FALSE = "false";
	
	/*
	public static void main(String args[]) {
//		Lexer lexer = new StringLexer("()");
//		for(int i=0; i<5; i++) {
//			Token tok = lexer.nextToken();
//			System.out.println(tok.code+":"+tok.info);
//		}
		ReadingTree tree = new ReadingTree("((a (c)) (b (d)))", new OutputStreamLog(System.out));
		System.out.println(tree.toString());
		System.out.println(tree.toString(0));
	}
	*/
	/**
	 * Removes a node from his father's list
	 */
	public static void removeSonFromFather(List forest, ReadingTreeNode node1) {
		ReadingTreeNode father = node1.father;
		if (father == null) {
			forest.remove(node1);
		} else {
			father.sons.remove(node1);
			node1.father = null;
		}
	}

	/*
	/** 
	 * Get the external representation of the tree
	*/
	public static String toString(List forest) {
		return "(" + makeForest(forest)+ ")";
	}
	public static String makeForest(List forest) {
		String res = "";
		for(int i=0; i<forest.size(); i++) {
			ReadingTreeNode node = (ReadingTreeNode)forest.get(i);
			res += makeTree(node);
		}
		return res;
	}
	public static String makeTree(ReadingTreeNode node) {
		String res = "(";
		res += node.info;
		List sons = node.sons;
		if (sons == null || sons.isEmpty()) ;
		else res += makeForest(sons);
		res += ")";
		return res;
	}
	
	
	/*
	// test program
    // ------------

    private static void pr(String s) {
      System.out.println(s);
    }

    private static void hd(String s) {
      System.out.println("\n" + s + "\n");
    }

    private static void pg() {
      try {
        int i;
        System.out.print("next page? (CR)");
        do i = System.in.read(); while (i != '\n');
        System.out.println();
      }
      catch (Exception e) {System.out.println("raised exception : " + e);}
    }

    public static void main(String args[]) throws Exception {
    	/*
    	{   		
    		hd("Tests of translate, computeTree, computeForest");
	    	pr("new ReadingTree(\")\")="+new ReadingTree(")"));
	    	pg();
	    	
	    	pr("new ReadingTree(\"(())\")="+new ReadingTree("(())"));
	    	pg();
	    	
	    	pr("new ReadingTree(\"()\")="+new ReadingTree("()"));
	    	pg();
	    	
	    	pr("new ReadingTree(\"(fred)\")="+new ReadingTree("(fred)"));
	    	pg();
	    	
	    	pr("new ReadingTree(\"(fred!\")="+new ReadingTree("(fred!"));
	    	pg();
	    	
	    	pr("new ReadingTree(\"(fred(jim)(joe))\")="+new ReadingTree("(fred(jim)(joe))"));
	    	pg();
	    	
       	}
       	
    	{
    		hd("Tests of getNodeNames");
	    	ReadingTree rt = new ReadingTree("");
	    	pr("rt="+rt);
	    	pr("rt.getNodeNames()"+rt.getNodeNames());
	    	pg();
	    	rt = new ReadingTree("(a(b))(c(d))");
	    	pr("rt"+rt);
	    	pr("rt.getNodeNames()="+rt.getNodeNames());
	    	pg();
    	}
    	/*
    	{
    		hd("Tests of find");
	    	ReadingTree rt = new ReadingTree("(test)");
    		pr("rt.find(\"fred\",null)="+rt.find("fred",null));
    		pg();
    		
    		pr("rt.find(\"fred\",new ReadingTreeNode(\"fred\"))="+
    			rt.find("fred",new ReadingTreeNode("fred")));
    		pg();
    		
    		pr("rt.find(\"fred\",new ReadingTreeNode(\"jim\"))="+
    			rt.find("fred",new ReadingTreeNode("jim")));
    		pg();
    		
    		ReadingTreeNode jimjackjoejohn = new ReadingTreeNode("jim");
    		ReadingTreeNode jack = new ReadingTreeNode("jack");
    		ReadingTreeNode joejohn = new ReadingTreeNode("joe");
    		ReadingTreeNode john = new ReadingTreeNode("john");
    		joejohn.addSon(john);
    		jimjackjoejohn.addSon(jack);
    		jimjackjoejohn.addSon(joejohn);
    		pr("jimjackjoejohn="+jimjackjoejohn);
    		pr("rt.find(\"fred\",jimjackjoejohn)="+
        		rt.find("fred",jimjackjoejohn));
        	pg();
        		
    		ReadingTreeNode jimjackjoefred = new ReadingTreeNode("jim");
    		ReadingTreeNode joefred = new ReadingTreeNode("joe");
    		ReadingTreeNode fred = new ReadingTreeNode("fred");
    		joefred.addSon(fred);
    		jimjackjoefred.addSon(jack);
    		jimjackjoefred.addSon(joefred);
    		pr("jimjackjoefred="+jimjackjoefred);
    		pr("rt.find(\"fred\",jimjackjoefred)="+
        		rt.find("fred",jimjackjoefred));
        	pg();
    	}
    	
    	{
    		hd("Tests of isAncestor, toDebugString");
	    	ReadingTree rt = new ReadingTree("(a(b))(c(d))");
	    	pr("rt="+rt);
	    	pr("rt.toDebugString()="+rt.toDebugString());
	    	pg();
	    	
	    	pr("rt.isAncestor(\"a\",\"a\")="+rt.isAncestor("a","a"));
	    	pr("rt.isAncestor(\"a\",\"b\")="+rt.isAncestor("a","b"));
	    	pr("rt.isAncestor(\"a\",\"c\")="+rt.isAncestor("a","c"));
	    	pr("rt.isAncestor(\"a\",\"d\")="+rt.isAncestor("a","d"));
	    	pr("rt.isAncestor(\"a\",\"e\")="+rt.isAncestor("a","e"));
	    	pg();
	    	pr("rt.isAncestor(\"b\",\"a\")="+rt.isAncestor("b","a"));
	    	pr("rt.isAncestor(\"b\",\"b\")="+rt.isAncestor("b","b"));
	    	pr("rt.isAncestor(\"b\",\"c\")="+rt.isAncestor("b","c"));
	    	pr("rt.isAncestor(\"b\",\"d\")="+rt.isAncestor("b","d"));
	    	pr("rt.isAncestor(\"b\",\"e\")="+rt.isAncestor("b","e"));
	    	pg();
	    	pr("rt.isAncestor(\"c\",\"a\")="+rt.isAncestor("c","a"));
	    	pr("rt.isAncestor(\"c\",\"b\")="+rt.isAncestor("c","b"));
	    	pr("rt.isAncestor(\"c\",\"c\")="+rt.isAncestor("c","c"));
	    	pr("rt.isAncestor(\"c\",\"d\")="+rt.isAncestor("c","d"));
	    	pr("rt.isAncestor(\"c\",\"e\")="+rt.isAncestor("c","e"));
	    	pg();
	    	pr("rt.isAncestor(\"d\",\"a\")="+rt.isAncestor("d","a"));
	    	pr("rt.isAncestor(\"d\",\"b\")="+rt.isAncestor("d","b"));
	    	pr("rt.isAncestor(\"d\",\"c\")="+rt.isAncestor("d","c"));
	    	pr("rt.isAncestor(\"d\",\"d\")="+rt.isAncestor("d","d"));
	    	pr("rt.isAncestor(\"d\",\"e\")="+rt.isAncestor("d","e"));
	    	pg();
	    	pr("rt.isAncestor(\"e\",\"a\")="+rt.isAncestor("e","a"));
	    	pr("rt.isAncestor(\"e\",\"b\")="+rt.isAncestor("e","b"));
	    	pr("rt.isAncestor(\"e\",\"c\")="+rt.isAncestor("e","c"));
	    	pr("rt.isAncestor(\"e\",\"d\")="+rt.isAncestor("e","d"));
	    	pr("rt.isAncestor(\"e\",\"e\")="+rt.isAncestor("e","e"));
	    	pg();
    	}
    	
    }
    */
}
