package rwd.rules.constraintdiagrams;

//import java.util.ArrayList;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.Vector;
//import java.util.Vector;

//import rwd.abstractsyntax.eulerdiagram.Contour;
import rwd.abstractsyntax.eulerdiagram.Zone;
import rwd.abstractsyntax.spiderdiagram.Spider;
import rwd.abstractsyntax.spiderdiagram.UnitaryDiagram;
import rwd.rules.CDRule;
import rwd.rules.application.RuleAppliesToSpiderAndZone;
import rwd.concretesyntax.*;
import rwd.concretesyntax.eulerdiagram.ConcreteZone;
import rwd.concretesyntax.spiderdiagram.ConcreteSpider;

/**
 * This is an adapter for the corresponding spider diagram rule 
 * in rwd.rules.choicespec.
 * 
 * This is a singleton class.  Use getRule to obtain the single instantiation.
 * 
 * @author Richard Bosworth
 * 
 * @version 24.8.05   20:13
 */
public class RuleAddSpiderFoot 
	extends CDRule 
{
	private rwd.rules.choicespec.RuleAddSpiderFoot mySDRule = 
		(rwd.rules.choicespec.RuleAddSpiderFoot) 
		rwd.rules.choicespec.RuleAddSpiderFoot.getRule();
	
	private Reconciler myReconciler = new Reconciler();
	
	
	// private constructor
	// -------------------	
	private RuleAddSpiderFoot() {
		super("cd: Add spider foot");
	}
	private static RuleAddSpiderFoot example = null;

	/**
	 * @return single instantiation of this class
	 */
	public static RuleAddSpiderFoot getRule() {
		if (example == null) {
			example = new RuleAddSpiderFoot();
		}
		return example;
	}
	
	// queries
	// -------

	/**
	 * @return true iff this rule needs extra arguments from the user
	 */
    public boolean needsExtraArgs() {
		return false;
	}
	
	/**
	 * Note: this rule needs a concrete unitary diagram, one or more zones, 
	 *   and an optional existential spider to be specified.
	 *   If no spider is specified, the diagram must have exactly one 
	 *     existential spider.
	 *   The specified zones must not be in the habitat of the existential spider. 
	 * @param args list of components to which this rule may apply
	 * @param extraArgs string representing extra arguments to which this rule may apply
	 * @return true iff this rule applies to the specified components and extra arguments
	 */
    /*
     * test:
     *   no concrete unitary diagram specified
     *   concrete unitary diagram specified
     *     no free zones in diagram
     *       one zone specified
     *     free zones in diagram
     *       no zones specified
     *       one zone specified
     *         two existential spiders selected
     *         one universal spider selected
     *         one existential spider selected
     *           specified zone is in habitat
     *           specified zone is not in habitat
     *             reading tree is not consistent
     *             reading tree is consistent
     *         no spider selected
     *           no existential spiders in diagram
     *           one existential spider in diagram
     *             specified zone is in habitat
     *             specified zone is not in habitat
     *           two existential spiders in diagram
     */
	public boolean canApplyTo(List args, String extraArgs) {
		ConcreteUnitaryDiagram cud = getCUD(args);
		if (cud == null) return false;
		UnitaryDiagram ud = cud.getAbstractUnitaryDiagram();
		List aspiders = ud.getSpiders();
		Vector azones = ud.getCloneOfZoneSet();

		if (!mySDRule.canApplyTo(ud)) return false;

		Collection zones = getConcreteZones(args);
		if (zones.isEmpty()) return false;
		
		Collection spiders = getConcreteSpiders(args);
		if (spiders.size()> 1) return false;
		
		// find required spider
		ConcreteSpider cs = null;
		Iterator sit = spiders.iterator();
		while (sit.hasNext()) {
			ConcreteSpider s = (ConcreteSpider) sit.next();
			if (!s.isExistential()) return false;
			else cs = s;
		}
		if (spiders.size()==0) {
			// user has not specified spiders - check there is just one in diagram
			int cardes = 0;  // cardinality of existential spiders
			Set cudspiders = cud.getConcreteSpiders();
			sit = cudspiders.iterator();
			while (sit.hasNext()){
				ConcreteSpider s = (ConcreteSpider) sit.next();
				if (s.isExistential()) {
					cardes++;
					cs = s;
				}
			}
			if (cardes!=1) return false;
		}
		// cs is the required spider
		
		// check specified zones are not in cs's habitat
		Set habitat = cs.getHabitat(cud);
		Iterator zit = zones.iterator();
		while (zit.hasNext()) {
			ConcreteZone cz = (ConcreteZone) zit.next();
			if (habitat.contains(cz)) return false;
		}
		//return true;
		
		// reading tree check not working at present 24.8.05 
		// check resultant diagram is consistent with reading tree
		ConcreteUnitaryDiagram newcud = applyRule(args,extraArgs);
		boolean result = newcud.isValidTree();
		System.out.println("canApplyTo.result="+result);
		return result;
	}

	
	// commands
	// --------
	
    /**
     * @param args list of components to which this rule must apply
     * @param extraArgs string representing extra arguments to which this rule must apply
     * @return diagram representing the result of applying this rule
     */
	/*
	 * test:
	 *   user selects two zones
	 *     user selects an existential spider
	 *     user selects no existential spider, but diagram contains one
	 */
	public ConcreteUnitaryDiagram applyRule(List args, String extraArgs) {
		ConcreteUnitaryDiagram cud = getCUD(args);
		
		// get selected zones
		Collection czones = getConcreteZones(args);
		Iterator czit = czones.iterator();
		Collection zones = new HashSet();
		while (czit.hasNext()){
			Zone z = ((ConcreteZone) czit.next()).getAbstractZone();
			Collection contours = z.getAllContours();
			zones.add(z);
		}
		
		// get spider
		Spider s;
		Collection argspiders = getConcreteSpiders(args);
		if (argspiders.size()==1) {
			Iterator it = argspiders.iterator();
			s = ((ConcreteSpider)it.next()).getAbstractSpider(cud);
		} else {
			Set cudspiders = cud.getConcreteSpiders();
			Iterator it = cudspiders.iterator();
			s = ((ConcreteSpider)it.next()).getAbstractSpider(cud);			
		}
		String spiderLabel = s.getLabel();
		
		UnitaryDiagram ud = cud.getAbstractUnitaryDiagram();
		
		// apply rule
		Collection apps = new ArrayList();
		Iterator zit = zones.iterator();
		while (zit.hasNext()){
			RuleAppliesToSpiderAndZone rasz = 
				new RuleAppliesToSpiderAndZone(ud,s,(Zone)zit.next());
			ud = (UnitaryDiagram) mySDRule.applyRule(rasz);
			if (zit.hasNext()) {
				// get new version of spider
				Collection spiders = ud.getSpiders();
				Iterator sit = spiders.iterator();
				while (sit.hasNext()) {
					Spider spider = (Spider) sit.next();
					if (spider.getLabel().equals(spiderLabel)) {
						s = spider;
						break;
					}
				}
			}
		}
		
		// we now have a possible abstract diagram ud corresponding to 
		//   one or more applications of the rule
		ConcreteUnitaryDiagram newcud = null;
		if (ud != null) {
		    newcud = myReconciler.reconcile(cud,ud);
		 }

		return newcud;
	}
	
	/*
	// test program
    // ------------
	
    private static void pr(String s) {
      System.out.println(s);
    }

    private static void hd(String s) {
      System.out.println("\n" + s + "\n");
    }

    private static void pg() {
      try {
        int i;
        System.out.print("next page? (CR)");
        do i = System.in.read(); while (i != '\n');
        System.out.println();
      }
      catch (Exception e) {System.out.println("raised exception : " + e);}
    }
    
    public static void main(String args[]) throws Exception {
    	
    	RuleAddSpiderFoot r = RuleAddSpiderFoot.getRule(); 
    	
    	{
	    	
	    	/*
	    	Contour a1 = Contour.getContour("a");
	    	Contour a2 = Contour.getContour("a");
	    	Contour b = Contour.getContour("b");
	    	pr("a1="+a1);
	    	pr("a2="+a2);
	    	pr("b="+b);
	    	pg();
	    	
	    	pr("(a1==a1)="+(a1==a1));
	    	pr("(a1==a2)="+(a1==a2));
	    	pr("(a1==b)="+(a1==b));
	    	pg();
	    	
	    	pr("(a1.equals(a1))="+(a1.equals(a1)));
	    	pr("(a1.equals(a2))="+(a1.equals(a2)));
	    	pr("(a1.equals(b))="+(a1.equals(b)));
	    	pg();
	    	
	    	Vector va1 = new Vector();
	    	va1.add(a1);
	    	
	    	Vector va2 = new Vector();
	    	va2.add(a2);
	    	
	    	Vector vb = new Vector();
	    	vb.add(b);
	    	
	    	Vector v = new Vector();
	    	
	    	Zone za1 = Zone.getZone(v,va1);
	    	Zone za2 = Zone.getZone(v,va2);
	    	Zone zb = Zone.getZone(v,vb);
	    	pr("za1="+za1);
	    	pr("za2="+za2);
	    	pr("zb="+zb);
	    	pg();
	    	
	    	pr("(za1==za1)="+(za1==za1));
	    	pr("(za1==za2)="+(za1==za2));
	    	pr("(za1==zb)="+(za1==zb));
	    	pg();
	    	
	    	pr("(za1.equals(za1))="+(za1.equals(za1)));
	    	pr("(za1.equals(za2))="+(za1.equals(za2)));
	    	pr("(za1.equals(zb))="+(za1.equals(zb)));
	    	pg();
	    	
    	}
    	
    }
    */
}
