package rwd.rules.constraintdiagrams;

//import java.awt.Color;
//import java.awt.Polygon;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;
//import java.util.Vector;

//import rwd.abstractsyntax.eulerdiagram.Contour;
import rwd.abstractsyntax.constraintdiagram.Arrow;
import rwd.abstractsyntax.eulerdiagram.Zone;
import rwd.abstractsyntax.spiderdiagram.Spider;
import rwd.abstractsyntax.spiderdiagram.UnitaryDiagram;
import rwd.concretesyntax.ConcreteUnitaryDiagram;
//import rwd.concretesyntax.eulerdiagram.ConcreteContour;
//import rwd.concretesyntax.eulerdiagram.DecorationBox;
import rwd.concretesyntax.spiderdiagram.ConcreteSpider;
//import rwd.concretesyntax.spiderdiagram.ConcreteSpiderFoot;
//import rwd.concretesyntax.spiderdiagram.ConcreteSpiderLeg;
import rwd.ocl.constraintdiagram.ReadingTree;
import rwd.ocl.constraintdiagram.ReadingTreeNode;
import rwd.ocl.constraintdiagram.ReadingTreeString;
import rwd.rules.CDRule;

/**
 * This is an adapter for the corresponding spider diagram rule 
 * in rwd.rules.choicespec.
 * 
 * This is a singleton class.  Use getRule to obtain the single instantiation.
 * <P>
 * N.B. This version of the rule asumes that a spider's domain is
 * the same as its habitat - this is not strictly true.
 * 
 * @author Richard Bosworth
 * 
 * @version 5.9.05   12:14
 */
public class RuleDeleteExistentialSpider extends CDRule {

	// constructor and accessor
	// ------------------------
	
	/**
	 * @param name
	 */
	private RuleDeleteExistentialSpider() {
		super("cd: Delete existential spider");
	}

	private static RuleDeleteExistentialSpider example = null;

	/**
	 * @return single instantiation of this class
	 */
	public static RuleDeleteExistentialSpider getRule() {
		if (example == null) {
			example = new RuleDeleteExistentialSpider();
		}
		return example;
	}
	
	// attributes
	// ----------
	
	/**
	 * @return true iff this rule needs extra arguments from the user
	 */
    public boolean needsExtraArgs() {
		return false;
	}
	
	/**
	 * Note: this rule needs a concrete unitary diagram, 
	 *   and optional existential spiders to be specified.
	 *   If no spider is specified, the diagram must have at least one 
	 *     existential spider.
	 * A spider s can only be deleted if none of the following conditions holds:
	 * <UL>
	 *   <LI>the habitat of s includes a shaded zone
	 *   <LI>s is the source or target of any arrow
	 *   <LI>the is a universal spider x whose domain includes a foot of the 
	 *        spider s and s is ordered before x in the reading tree
	 * </UL>
	 * @param args list of components to which this rule may apply
	 * @param extraArgs string representing extra arguments to which this rule may apply
	 * @return true iff this rule applies to the specified components and extra arguments
	 */
    /*
     * post: spiderSet = set of abstract spiders to which this rule is to be applied
     *                     (may be empty)
     */
    /*
     * test:
     *   no concrete unitary diagram specified
     *   concrete unitary diagram specified
     *     specified spider's habitat has shaded zones (to trip SDRule)
     *     specified spider is universal
     *     specified spider is existential
     *       specified spider is source of arrow
     *       specified spider has universal descendant in reading tree which is in same zone
     *       specified spider has no shaded zones, no connected arrow, no universal descendant
     *     no spider specified
     *     no existential spiders in diagram
     *     one or more existential spiders in diagram, 
     *       all of which are deletable
     */
	public boolean canApplyTo(List args, String extraArgs) {
		spiderSet = new LinkedHashSet();
		ConcreteUnitaryDiagram cud = getCUD(args);
		if (cud == null) return false;
		UnitaryDiagram ud = cud.getAbstractUnitaryDiagram();
		Collection aspiders = ud.getSpiders();
		//Vector azones = ud.getCloneOfZoneSet();

		if (!mySDRule.canApplyTo(ud)) return false;

		Collection argSpiders = getConcreteSpiders(args);
		
		// find required spiders
		Iterator sit = argSpiders.iterator();
		while (sit.hasNext()) {
			Spider s = ((ConcreteSpider) sit.next()).getAbstractSpider(cud);
			if (!s.isExistential()) return false;
			else spiderSet.add(s);
		}
		if (spiderSet.size()==0) {
			// user has not specified spiders - check there is at least one in diagram
			sit = aspiders.iterator();
			while (sit.hasNext()){
				Spider s = (Spider) sit.next();
				if (s.isExistential()) {
					spiderSet.add(s);
				}
			}
			if (spiderSet.size()<1) return false;
		}
		// spiderSet has the putative spiders
		
		// for each spider s in set
		sit = spiderSet.iterator();
		while (sit.hasNext()) {
			Spider s  = (Spider) sit.next();
			
			// check s's zones not shaded
			if (hasShadedHabitat(s,ud)) {
				spiderSet.clear();
				break;
			}
		
			// check s not source or target of arrow
			if (isSourceOrTarget(s,ud)) {
				spiderSet.clear();
				break;
			}
		
			// check no universal spider after s in the tree has s in its habitat
			if (descendantsDomainIncludes(s,new ReadingTree(ud.getReadingForest()),ud)) {
				spiderSet.clear();
				break;
			}
		}
		return ! spiderSet.isEmpty();
	}

	
	// operations
	// ----------
	
	
    /**
     * pre:  canApplyTo must have been executed on the same arguments 
     * @param args list of components to which this rule must apply
     * @param extraArgs string representing extra arguments to which this rule must apply
     * @return diagram representing the result of applying this rule
     */
    /*
     * pre:  spiderSet = set of abstract spiders to which this rule is to be applied
     *                     (may be empty)
     */
 	/*
	 * test:
	 *   no spiders in spiderSet
	 *   two spiders in spiderSet
	 */
	public ConcreteUnitaryDiagram applyRule(List args, String extraArgs) {
		ConcreteUnitaryDiagram cud = getCUD(args);
		UnitaryDiagram ud = cud.getAbstractUnitaryDiagram();
		ReadingTreeString rts = new ReadingTreeString(ud.getReadingForest());

		// iterate though spiders, deleting them from abstract diagram and tree
		Iterator sit = spiderSet.iterator();
		while (sit.hasNext()) {
			Spider s = (Spider)sit.next();
			ud.removeSpider(s);
			rts.deleteNode(s.getLabel());			
		}
		ud.setReadingForest(rts.toString());
		
		// we now have a possible abstract diagram ud corresponding to 
		//   one or more applications of the rule
		ConcreteUnitaryDiagram newcud = null;
		if (ud != null) {
		    newcud = myReconciler.reconcile(cud,ud);
		 }

		return newcud;
	}
	
	// private variables
	// -----------------
	private rwd.rules.choicespec.RuleDeleteSpider mySDRule = 
		(rwd.rules.choicespec.RuleDeleteSpider) 
		rwd.rules.choicespec.RuleDeleteSpider.getRule();
	
	private Reconciler myReconciler = new Reconciler();
	
	private Set spiderSet;  // set of spiders on which this rule is to be performed
	
	
	// private methods
	// ---------------
	
	/*
	 * pre:  s = abstract spider
	 *       ud = unitary diagram containing s
	 * post: returns true iff s has one or more shaded zones in its habitat
	 */
	/*
	 * test:
	 *   s has two or more zones in habitat
	 *     one shaded zone in habitat
	 *     no shaded zones in habitat
	 */
	private boolean hasShadedHabitat(Spider s, UnitaryDiagram ud) {
		Collection h = s.getHabitat();
		Iterator hit = h.iterator();
		while (hit.hasNext()) {
			Zone z = (Zone) hit.next();
			if (z.isShaded(ud)) return true;
		}
		return false;
	}


	/*
	 * pre:  s = abstract spider
	 *       ud = unitary diagram containing s
	 * post: returns true iff s is the source or target of one or more arrows
	 */
	/*
	 * test:
	 *   ud has no arrows
	 *   ud has two or more arrows
	 *     s is source of an arrow
	 *     s is target of an arrow
	 *     s is not source or target of an arrow
	 */
	private boolean isSourceOrTarget(Spider s, UnitaryDiagram ud) {
		
		Collection arrows = ud.getArrows();
		Iterator ait = arrows.iterator();
		while (ait.hasNext()) {
			Arrow a = (Arrow) ait.next();
			if (a.getSource() == s) return true;
			if (a.getTarget() == s) return true;
		}
		return false;
	}
	
	/*
	 * pre:  s = spider whose descendants in rts are to be tested
	 *       rt = reading tree from which descendants of s are to be tested
	 *       ud = unitary diagram of which s is a member
	 * post: returns true iff there exists a universal spider x, which is 
	 *         a descendant of s, and which has a foot of s in its domain
	 * NB. This version of the method assumes that a spider's domain
	 *     is the same as its habitat - not strictly true.
	 */
	/*
	 * test:
	 *   s not in reading tree
	 *   s in reading tree
	 *     s has no descendants
	 *     s has descendant that is not in ud
	 *     s has descendant x that is in ud
	 *       x is not universal
	 *       x is universal
	 *         x has habitat not containing foot of s
	 *         x has habitat containing foot of s
	 */
	private boolean descendantsDomainIncludes(Spider s, 
			                                  ReadingTree rt, 
											  UnitaryDiagram ud) {
		ReadingTreeNode rtn = rt.find(s.getLabel());
		if (rtn == null) return false;
		Iterator dit = rtn.getDescendants().iterator();
		while (dit.hasNext()) {
			String name = ((ReadingTreeNode) dit.next()).getInfo();
			Spider sp = ud.getSpiderCalled(name);
			if (sp == null) return false;
			if (!sp.isUniversal()) return false;
			Iterator hit = sp.getHabitat().iterator();
			while (hit.hasNext()) {
				Zone z = (Zone) hit.next();
				Iterator h2it = s.getHabitat().iterator();
				while (h2it.hasNext()) {
					Zone z2 = (Zone) h2it.next();
					if (z2.equals(z))
						return true;
				}
			}
		}
		return false;
	}

	/*
	// test program
    // ------------

    private static void pr(String s) {
      System.out.println(s);
    }

    private static void hd(String s) {
      System.out.println("\n" + s + "\n");
    }

    private static void pg() {
      try {
        int i;
        System.out.print("next page? (CR)");
        do i = System.in.read(); while (i != '\n');
        System.out.println();
      }
      catch (Exception e) {System.out.println("raised exception : " + e);}
    }

    public static void main(String args[]) throws Exception {
    	RuleDeleteExistentialSpider rdes = RuleDeleteExistentialSpider.getRule();
    	pr("rdes="+rdes);
    	pr("rdes.hashCode()="+rdes.hashCode());
    	
    	RuleDeleteExistentialSpider rdes2 = RuleDeleteExistentialSpider.getRule();
    	pr("rdes2.hashCode()="+rdes2.hashCode());
    	pg();
    	/*
    	{
    		hd("Tests of hasShadedHabitat");
    		
    		Polygon dp1 = new Polygon();
    		dp1.addPoint(0,0);
    		dp1.addPoint(20,0);
    		dp1.addPoint(20,20);
    		dp1.addPoint(0,20);
    		DecorationBox db1 = new DecorationBox("db1","",dp1);
    		pr("db1="+db1);
    		
    		Polygon p1 = new Polygon();
    		p1.addPoint(10,10);
    		p1.addPoint(10,15);
    		p1.addPoint(15,15);
    		p1.addPoint(15,10);
    		ConcreteContour cc1 = new ConcreteContour(Contour.getContour("a"),p1);
    		pr("cc1="+cc1);
    		
    		ConcreteSpiderFoot csf1 = 
    			 new ConcreteSpiderFoot("s",ConcreteSpiderFoot.NORMAL);
    		csf1.setLocation(12,12);
    		pr("csf1="+csf1);
    		ConcreteSpiderFoot csf2 = 
    			new ConcreteSpiderFoot("",ConcreteSpiderFoot.NORMAL);
    		csf2.setLocation(5,5);
    		pr("csf2="+csf2);
    		
    		ConcreteSpiderLeg csl = 
    			new ConcreteSpiderLeg("",csf1,csf2,ConcreteSpiderLeg.NORMAL);
    		pr("csl="+csl);
    		
    		ConcreteUnitaryDiagram cud = new ConcreteUnitaryDiagram(db1);
    		cud.addConcreteContour(cc1);
    		cud.addConcreteFoot(csf1);
    		cud.addConcreteFoot(csf2);
    		cud.addConcreteLeg(csl);
    		pr("cud="+cud);
    		pg();
    		
    		UnitaryDiagram ud = cud.getAbstractUnitaryDiagram();
    		pr("ud="+ud.toDebugString());
    		
    		Collection ss = ud.getSpiders();
    		Iterator sit = ss.iterator();
    		Spider s = (Spider) sit.next();
    		pr("s="+s);
    		pg();
    		
    		pr("rdes.hasShadedHabitat(s,ud)=" +
    			rdes.hasShadedHabitat(s,ud));
    		pg();
    		
    		ConcreteZone cz1 = cud.getConcreteZone(csf1);
    		cz1.setBackColor(new Color(128,128,128));
    		pr("cud="+cud);
    		ud = cud.getAbstractUnitaryDiagram();
    		pr("ud="+ud.toDebugString());
    		pr("s="+s);
    		pg();
    		
    		pr("rdes.hasShadedHabitat(s,ud)=" +
        			rdes.hasShadedHabitat(s,ud));
       		pg();
    	}
    	*/
    	/*
    	{
    		hd("Tests of isSourceOrTarget");
    		Vector h = new Vector();
    		Spider s = new Spider("s",h);
    		pr("s="+s);
    		UnitaryDiagram ud = new UnitaryDiagram();
    		ud.addSpider(s);
    		pr("ud="+ud.toDebugString());
    		pg();
    		
    		pr("rdes.isSourceOrTarget(s,ud)="+rdes.isSourceOrTarget(s,ud));
    		pg();
    		
    		Arrow a = new Arrow(s,null,"a");
    		Arrow b = new Arrow(null,null,"b");
    		ud.addArrow(a);
    		ud.addArrow(b);
    		pr("ud="+ud.toDebugString());
    		pr("rdes.isSourceOrTarget(s,ud)="+rdes.isSourceOrTarget(s,ud));
    		pg();
    		
    		a = new Arrow(null,null,"a");
    		b = new Arrow(null,s,"b");
    		ud = new UnitaryDiagram();
    		ud.addSpider(s);
    		ud.addArrow(a);
    		ud.addArrow(b);
    		pr("ud="+ud.toDebugString());
    		pr("rdes.isSourceOrTarget(s,ud)="+rdes.isSourceOrTarget(s,ud));
    		pg();
    		
    		a = new Arrow(null,null,"a");
    		b = new Arrow(null,null,"b");
    		ud = new UnitaryDiagram();
    		ud.addSpider(s);
    		ud.addArrow(a);
    		ud.addArrow(b);
    		pr("ud="+ud.toDebugString());
    		pr("rdes.isSourceOrTarget(s,ud)="+rdes.isSourceOrTarget(s,ud));
    		pg();    		
    	}
    	*/
    	/*
    	{
    		hd("Tests of descendantsDomainIncludes");
    		UnitaryDiagram ud = new UnitaryDiagram();
    		pr("ud="+ud.toDebugString());
    		ReadingTree rt = new ReadingTree("(a)(b(c))");
    		pr("rt="+rt);
    		Vector h = new Vector();
    		Spider s = new Spider("d",h);
    		pr("s="+s);
    		pr("s.kind="+s.kind);
    		pg();
    		
    		pr("rdes.descendantsDomainIncludes(s,rt,ud)=" +
        			rdes.descendantsDomainIncludes(s,rt,ud));
        	pg();
       		
        	s = new Spider("a",h);
    		pr("s="+s);
    		pr("s.kind="+s.kind);
        	pr("rdes.descendantsDomainIncludes(s,rt,ud)=" +
            	rdes.descendantsDomainIncludes(s,rt,ud));
            pg();
           		
        	s = new Spider("b",h);
    		pr("s="+s);
    		pr("s.kind="+s.kind);
        	pr("rdes.descendantsDomainIncludes(s,rt,ud)=" +
            	rdes.descendantsDomainIncludes(s,rt,ud));
            pg();
            
            Contour a = Contour.getContour("a");
            pr("a="+a);
            Vector conta = new Vector();
            conta.add(a);
            Vector exclnil = new Vector();
            Zone ina = new Zone(conta,exclnil);
            pr("ina="+ina);
            
            Vector contnil = new Vector();
            Vector excla = new Vector();
            excla.add(a);
            Zone outa = new Zone(contnil,excla);
            pr("outa="+outa);

            Vector shabitat = new Vector();
            shabitat.add(ina);
            s = new Spider("b",shabitat);
            pr("s="+s);
    		pr("s.kind="+s.kind);
            
            Vector thabitat = new Vector();
            thabitat.add(ina);
            Spider t = new Spider("c",thabitat);
            pr("t="+t);
    		pr("t.kind="+t.kind);
    		Vector zones = new Vector();
    		zones.add(ina);
    		zones.add(outa);
    		ud.setNewZoneSet(zones);
    		ud.addSpider(s);
    		ud.addSpider(t);
    		pr("ud="+ud.toDebugString());
            pr("s="+s);
    		pr("s.kind="+s.kind);
            pr("t="+t);
    		pr("t.kind="+t.kind);
        	pr("rdes.descendantsDomainIncludes(s,rt,ud)=" +
                rdes.descendantsDomainIncludes(s,rt,ud));
            pg();
                
            thabitat = new Vector();
            thabitat.add(outa);
            t = new Spider("c",thabitat);
            t.kind = Spider.UNIVERSAL;
            ud = new UnitaryDiagram();
    		ud.setNewZoneSet(zones);
    		ud.addSpider(s);
    		ud.addSpider(t);
    		pr("ud="+ud.toDebugString());
            pr("s="+s);
    		pr("s.kind="+s.kind);
            pr("t="+t);
    		pr("t.kind="+t.kind);
        	pr("rdes.descendantsDomainIncludes(s,rt,ud)=" +
                rdes.descendantsDomainIncludes(s,rt,ud));
            pg();
                
            thabitat = new Vector();
            thabitat.add(ina);
            t = new Spider("c",thabitat);
            t.kind = Spider.UNIVERSAL;
            ud = new UnitaryDiagram();
    		ud.setNewZoneSet(zones);
    		ud.addSpider(s);
    		ud.addSpider(t);
    		pr("ud="+ud.toDebugString());
            pr("s="+s);
    		pr("s.kind="+s.kind);
            pr("t="+t);
    		pr("t.kind="+t.kind);
        	pr("rdes.descendantsDomainIncludes(s,rt,ud)=" +
                rdes.descendantsDomainIncludes(s,rt,ud));
            pg();
    	}
    	*/
    	/*
    }
    */
	
}
