package rwd.rules.constraintdiagrams;

import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;

//import rwd.abstractsyntax.eulerdiagram.Contour;
import rwd.abstractsyntax.eulerdiagram.Zone;
import rwd.abstractsyntax.spiderdiagram.UnitaryDiagram;
import rwd.rules.CDRule;
import rwd.rules.application.RuleAppliesToZone;
import rwd.concretesyntax.*;
import rwd.concretesyntax.eulerdiagram.ConcreteZone;

/**
 * This is an adapter for the corresponding spider diagram rule 
 * in rwd.rules.choicespec.
 * 
 * This is a singleton class.  Use getRule to obtain the single instantiation.
 * 
 * @author Richard Bosworth
 * 
 * @version 8.8.05   17:05
 */
public class RuleUnshadeZone 
	extends CDRule 
{
	private rwd.rules.choicespec.RuleUnshadeZone mySDRule = 
		(rwd.rules.choicespec.RuleUnshadeZone) 
		rwd.rules.choicespec.RuleUnshadeZone.getRule();
	
	private Reconciler myReconciler = new Reconciler();
	
	
	// private constructor
	// -------------------	
	private RuleUnshadeZone() {
		super("cd: Erase Shading");
	}
	private static RuleUnshadeZone example = null;

	/**
	 * @return single instantiation of this class
	 */
	public static RuleUnshadeZone getRule() {
		if (example == null) {
			example = new RuleUnshadeZone();
		}
		return example;
	}
	
	// queries
	// -------

	/**
	 * @return true iff this rule needs extra arguments from the user
	 */
    public boolean needsExtraArgs() {
		return false;
	}
	
	/**
	 * @param args list of components to which this rule may apply
	 * @param extraArgs string representing extra arguments to which this rule may apply
	 * @return true iff this rule applies to the specified components and extra arguments
	 */
    /*
     * test:
     *   invalid diagram (no concrete unitary diagram)
     *   args from diagram with shaded zone
     *   args from diagram with no shaded zone
     */
	public boolean canApplyTo(List args, String extraArgs) {
		ConcreteUnitaryDiagram cud = getCUD(args);
		if (cud == null) return false;
		UnitaryDiagram ud = cud.getAbstractUnitaryDiagram();
		return mySDRule.canApplyTo(ud);
	}

	
	// commands
	// --------
	
    /**
     * @param args list of components to which this rule must apply
     * @param extraArgs string representing extra arguments to which this rule must apply
     * @return diagram representing the result of applying this rule
     */
	/*
	 * test:
	 *   user selects no zones
	 *     no application possibilities (i.e. no shaded zones)
	 *     two application possibilities (i.e. two shaded zones)
	 *   user selects two zones
	 */
	public ConcreteUnitaryDiagram applyRule(List args, String extraArgs) {
		ConcreteUnitaryDiagram cud = getCUD(args);
		Collection czones = getConcreteZones(args);
		Iterator czit = czones.iterator();
		Collection zones = new HashSet();
		while (czit.hasNext()){
			Zone z = ((ConcreteZone) czit.next()).getAbstractZone();
			Collection contours = z.getAllContours();
			zones.add(z);
		}
		UnitaryDiagram ud = cud.getAbstractUnitaryDiagram();
		
		Collection apps = mySDRule.getApplicationPossibilities(ud);
		Iterator appsit = apps.iterator();
		
		UnitaryDiagram rud = ud;
		while (appsit.hasNext()){
			RuleAppliesToZone raz = (RuleAppliesToZone) appsit.next();
			RuleAppliesToZone newraz = new RuleAppliesToZone(rud,raz.z);
			if (zones.isEmpty()){
				rud = (UnitaryDiagram) mySDRule.applyRule(newraz);
				break;  // only apply the rule once (to a random shaded zone!)
			}
			else{
				if (zones.contains(newraz.z)){
					rud = (UnitaryDiagram) mySDRule.applyRule(newraz);
				}
			}
		}
		
		// we now have a possible abstract diagram rud corresponding to 
		//   one or more applications of the rule
		ConcreteUnitaryDiagram newcud = null;
		if (rud != null) {
		    newcud = myReconciler.reconcile(cud,rud);
		 }

		return newcud;
	}
	
	
	// test program
    // ------------

	/*
	 * pre:  contours = collection of contours to be checked for rogues
	 * post: rogue contour report has been output on console
	 * 			with a bell for each rogue contour
	 */
	/*
	 * test:
	 *   empty collection
	 *   collection with at least two contours
	 *     one rogue contour
	 */
	/*
	private void checkRogueContours(Collection contours){
		pr("Rogue contours:");
		Iterator cit = contours.iterator();
		while(cit.hasNext()){
			Contour c = (Contour) cit.next();
			if (!Contour.isInContourLibrary(c))
				pr("  c="+c+" c.hashCode()="+c.hashCode()+"\7");
		}
		pr("");
	}
	*/
	/*
	 * pre:  zones = collection of zones to be checked for rogues
	 * post: rogue zone report has been output on console
	 * 			with a bell for each rogue zone
	 */
	/*
	 * test:
	 *   empty collection
	 *   collection with at least two zones
	 *     one rogue zone
	 */
	/*
	private void checkRogueZones(Collection zones){
		pr("Rogue zones:");
		Iterator zit = zones.iterator();
		while(zit.hasNext()){
			Zone z = (Zone) zit.next();
			if(!Zone.isInZoneLibrary(z))
				pr("  z="+z+" z.hashCode()="+z.hashCode()+"\7");
		}
		pr("");
	}
	
    private static void pr(String s) {
      System.out.println(s);
    }

    private static void hd(String s) {
      System.out.println("\n" + s + "\n");
    }

    private static void pg() {
      try {
        int i;
        System.out.print("next page? (CR)");
        do i = System.in.read(); while (i != '\n');
        System.out.println();
      }
      catch (Exception e) {System.out.println("raised exception : " + e);}
    }
    /*
    public static void main(String args[]) throws Exception {
    	
    	RuleUnshadeZone r = RuleUnshadeZone.getRule(); 
    	/*
    	{
	    	hd("Tests of contour and zone equality");
	    	Contour a1 = Contour.getContour("a");
	    	Contour a2 = Contour.getContour("a");
	    	Contour b = Contour.getContour("b");
	    	pr("a1="+a1);
	    	pr("a2="+a2);
	    	pr("b="+b);
	    	pg();
	    	
	    	pr("(a1==a1)="+(a1==a1));
	    	pr("(a1==a2)="+(a1==a2));
	    	pr("(a1==b)="+(a1==b));
	    	pg();
	    	
	    	pr("(a1.equals(a1))="+(a1.equals(a1)));
	    	pr("(a1.equals(a2))="+(a1.equals(a2)));
	    	pr("(a1.equals(b))="+(a1.equals(b)));
	    	pg();
	    	
	    	Vector va1 = new Vector();
	    	va1.add(a1);
	    	
	    	Vector va2 = new Vector();
	    	va2.add(a2);
	    	
	    	Vector vb = new Vector();
	    	vb.add(b);
	    	
	    	Vector v = new Vector();
	    	
	    	Zone za1 = Zone.getZone(v,va1);
	    	Zone za2 = Zone.getZone(v,va2);
	    	Zone zb = Zone.getZone(v,vb);
	    	pr("za1="+za1);
	    	pr("za2="+za2);
	    	pr("zb="+zb);
	    	pg();
	    	
	    	pr("(za1==za1)="+(za1==za1));
	    	pr("(za1==za2)="+(za1==za2));
	    	pr("(za1==zb)="+(za1==zb));
	    	pg();
	    	
	    	pr("(za1.equals(za1))="+(za1.equals(za1)));
	    	pr("(za1.equals(za2))="+(za1.equals(za2)));
	    	pr("(za1.equals(zb))="+(za1.equals(zb)));
	    	pg();
    	}
    
       	{
    		hd("Test of checkRogueContours");
    		Collection contours = new java.util.ArrayList();
    		pr("contours="+contours);
    		r.checkRogueContours(contours);
    		pg();
    		
    		Contour a = Contour.getContour("a");
    		Contour b = Contour.getContour("b");
    		Contour c = new Contour("c");
    		contours.add(a);
    		contours.add(b);
    		contours.add(c);
    		pr("contours="+contours);
    		r.checkRogueContours(contours);
    		pg();
       	}
       	*/
    	/*
       	{
    		hd("Test of checkRogueZones");
    		Collection zones = new java.util.ArrayList();
    		pr("zones="+zones);
    		r.checkRogueZones(zones);
    		pg();
    		
    		Vector va = new Vector();
    		Contour a = Contour.getContour("a");
    		va.add(a);
    		pr("va="+va);
    		
    		Vector vb = new Vector();
    		Contour b = Contour.getContour("b");
    		vb.add(b);
    		pr("vb="+vb);
    		
    		Vector v = new Vector();
    		pr("v="+v);
    		
    		Zone azb = Zone.getZone(va,vb);
    		Zone bza = Zone.getZone(vb,va);
    		Zone az = new Zone(va,v);
    		zones.add(azb);
    		zones.add(bza);
    		zones.add(az);
    		pr("zones="+zones);
    		r.checkRogueZones(zones);
    		pg();

    	}
       	
    }
    */

}
