package rwd.tableaux.diagram;

import java.util.Vector;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;

import rwd.abstractsyntax.eulerdiagram.Zone;
import rwd.abstractsyntax.spiderdiagram.AndCompoundDiagram;
import rwd.abstractsyntax.spiderdiagram.CompoundDiagram;
import rwd.abstractsyntax.spiderdiagram.Diagram;
import rwd.abstractsyntax.spiderdiagram.FalseDiagram;
import rwd.abstractsyntax.spiderdiagram.NotCompoundDiagram;
import rwd.abstractsyntax.spiderdiagram.OrCompoundDiagram;
import rwd.abstractsyntax.spiderdiagram.Spider;
import rwd.abstractsyntax.spiderdiagram.TrueDiagram;
import rwd.abstractsyntax.spiderdiagram.UnitaryDiagram;
import rwd.tableaux.semantic.Contradiction;
import rwd.tableaux.semantic.ContradictionImpl;
import rwd.tableaux.semantic.Options;
import uk.ac.kent.cs.kmf.util.ILog;

/**
 * @author Octavian Patrascoiu
 *
 */
public class CheckerImpl implements Checker {
	/** Constructors */
	public CheckerImpl(Options opt, ILog log) {
		this.opt = opt;
		this.log = log;
	}

	/** Computes the contradiction for a diagram: must be in DNF */
	public Contradiction computeContradiction(Diagram d) {
		Contradiction con = new ContradictionImpl();
		// true or not false
		if (isTrue(d) || isNotFalse(d)) {
			con.setContradiction(false);
			return con;
		//	false or not true
		} else if (isFalse(d) || isNotTrue(d)) { 
			con.setContradiction(true);
			return con;
		// Unitary diagram
		} else if (d.getClass() == UnitaryDiagram.class) {
			con.setContradiction(false);
			return con;
		// Compound diagram 			
		} else {
			d = transform(d);
			return scan4Cont(d);
		}
	}

	/** Check if there is a contradiction between ud1 and ud2 (both are unitary diagrams) */
	public Contradiction computeContradiction(Diagram d1, Diagram d2) {
		// Simple cases
		Contradiction con = new ContradictionImpl();
		// false
		if (isFalse(d1) || isFalse(d2) || isNotTrue(d1) || isNotTrue(d2)) {
			con.setContradiction(true);
			return con;
		}
		// true or not false
		if (isTrue(d1) || isTrue(d2) || isNotFalse(d1) || isNotFalse(d2)) {
			con.setContradiction(false);
			return con;
		}
		// d1 equals to d2
		if (d1.equals(d2)) {
			con.setContradiction(false);
			return con;			
		}
		// d1 is unitary, d2 is not d2', and d1 = d2'
		if (d1 instanceof UnitaryDiagram) {
			if (d2 instanceof NotCompoundDiagram) {
				Diagram d2p = (Diagram)((NotCompoundDiagram)d2).getChildren().get(0);
				if (d1.equals(d2p)) {
					con.setContradiction(false);
					return con;			
				}
			}
		}
		// d2 is unitary, d1 is not d1', and d2 = d1'
		if (d2 instanceof UnitaryDiagram) {
			if (d1 instanceof NotCompoundDiagram) {
				Diagram d1p = (Diagram)((NotCompoundDiagram)d1).getChildren().get(0);
				if (d2.equals(d1p)) {
					con.setContradiction(false);
					return con;			
				}
			}
		}
		// d1 and d2 are unitary diagrams
		if (d1 instanceof UnitaryDiagram && d2 instanceof UnitaryDiagram) {
			Set zones = computeContradictionSet((UnitaryDiagram)d1, (UnitaryDiagram)d2);
			if (zones.size() != 0) {
				con.setContradiction(true);
				con.setZones(zones);
			}
			return con; 
		}
		
		// Other cases
		// Transform diagrams
		d1 = transform(d1);
		d2 = transform(d2);
		// Make an and
		AndCompoundDiagram and = new AndCompoundDiagram(new Vector(), "");
		and.getChildren().add(d1);
		and.getChildren().add(d2);
		// Transform the and diagram
		Diagram d = transform(and);
		return computeContradiction(and);
	}

	/** Scan recursively the diagram */
	protected Contradiction scan4Cont(Diagram d) {
		Contradiction con = new ContradictionImpl();
		// null
		if (d == null) {
			con.setContradiction(false);
			return con;
		} 
		// true
		if (d instanceof TrueDiagram) { 
			con.setContradiction(false);
			return con;
		// false
		} else if (d instanceof FalseDiagram) {
			con.setContradiction(true);
			return con;
		// user unitary diagram
		} else if (d instanceof UnitaryDiagram ){
			con.setContradiction(false);
			return con;
		// Compound diagram
		} else if (d instanceof CompoundDiagram) {
			CompoundDiagram cd = (CompoundDiagram)d;
			List children = cd.getChildren();
			// not			
			if (d instanceof NotCompoundDiagram) {
				Diagram child = (Diagram)children.get(0);
				return computeContradiction(child).not();
			// And : all children are unitary diagrams
			} else if (d instanceof AndCompoundDiagram) {
				// Check every pair of children
				for(int i=0; i<children.size(); i++) {
					Diagram di = (Diagram)children.get(i);  
					for(int j=i+1; j<children.size(); j++) {
						Diagram dj = (Diagram)children.get(j);
						Contradiction temp = computeContradiction((UnitaryDiagram)di, (UnitaryDiagram)dj);
						if (temp.isContradiction()) {
							con.setContradiction(true);
							con.getZones().addAll(temp.getZones()); 
						}
					}
				}
				return con;
			// or
			} else if (d instanceof OrCompoundDiagram) {
				// Check all the children
				for(int i = 0; i<children.size(); i++) {
					Diagram child = (Diagram)children.get(i);
					Contradiction temp = scan4Cont(child);
					if (!temp.isContradiction()) {
						return new ContradictionImpl();
					} else {
						con.setContradiction(true);
						con.getZones().addAll(temp.getZones());						
					}
				}
				return con;
			// other compound diagram
			} else {
				return con;
			}
		// else
		} else {
			return con;
		}
	}

	/** Check if a diagram contains a contradiction: must be in DNF */
	protected Set computeContradictionSet(Diagram d) {
		// null
		if (d == null)
			return new TreeSet();
		// Unitary diagram
		// true
		if (d instanceof TrueDiagram) { 
			return new TreeSet();
		// false
		} else if (d instanceof FalseDiagram) {
			return new TreeSet();
		// atomic
		} else if (d instanceof UnitaryDiagram) {
			return new TreeSet();
		// Compound diagram
		} else if (d instanceof CompoundDiagram) {
			CompoundDiagram cd = (CompoundDiagram)d;
			List children = cd.getChildren();
			// not			
			if (d instanceof NotCompoundDiagram) {
				Diagram child = (Diagram)children.get(0);
				return computeContradictionSet(d);
			// And : all children are unitary diagrams
			} else if (d instanceof AndCompoundDiagram) {
				// Check every pair of children
				Set result =  new TreeSet();
				for(int i=0; i<children.size(); i++) {
					Diagram di = (Diagram)children.get(i);  
					for(int j=i+1; j<children.size(); j++) {
						Diagram dj = (Diagram)children.get(j);
						Set contr = computeContradictionSet((UnitaryDiagram)di, (UnitaryDiagram)dj);
						result.addAll(contr); 
					}
				}
				return result;
			// or
			} else if (d instanceof OrCompoundDiagram) {
				// Check all the children
				Set result = new TreeSet();
				for(int i = 0; i<children.size(); i++) {
					Diagram child = (Diagram)children.get(i);
					Set contr = computeContradictionSet(child);
					if (contr.size() == 0) {
						return contr;
					} else {
						result.addAll(contr);
					}
				}
				return result;
			// other compound diagram
			} else {
				return new TreeSet();
			}
		// else
		} else {
			return new TreeSet();
		}
	}
		
	/** Compute the zones with contrsdictions: ud1 and ud2 are user unitary diagrams */
	protected Set computeContradictionSet(UnitaryDiagram ud1, UnitaryDiagram ud2) { 
		// Store all the zones with contradictions
		Set result = new TreeSet();
		// Check all the zones
		Set contours = computeContours(ud1, ud2);
		List vennZones = UnitaryDiagram.allPossibleZones(new Vector(contours));
		List zones1 = ud1.getCloneOfZoneSet();
		List zones2 = ud2.getCloneOfZoneSet();
		for(int i=0; i<vennZones.size(); i++) {
			Zone zone = (Zone)vennZones.get(i);
			// Compute isShaded and noFeet for both diagrams
			boolean isShaded1 = isShaded(zone, ud1, zones1);
			boolean isShaded2 = isShaded(zone, ud2, zones2);
			int noFeet1 = computeNoFeet(zone, ud1);
			int noFeet2 = computeNoFeet(zone, ud2);
			
			// Check if there is a contradiction
			boolean flag = isContradiction(isShaded1, noFeet1, isShaded2, noFeet2); 
			if (flag) {
				result.add(zone);
			}
			
			// Debug info
			if (opt.isDebug()) {
				log.reportMessage("Check zone = "+zone);
				log.reportMessage("isShaded1(z) = "+isShaded1+" noFeet1(z)= "+noFeet1);
				log.reportMessage("isShaded2(z) = "+isShaded2+" noFeet2(z)= "+noFeet2);
				if (flag) log.reportMessage("CONTRADICTION");
				else log.reportMessage("OK");
			}
		}
		return result;
	}
	
	/** Transform the diagram */
	protected Diagram transform(Diagram d) {
		// Equalise the diagram
		Equaliser eq = new EqualiserImpl(opt, log);
		d = eq.equaliseDiagram(d);

		// Split the spiders
		SpiderSplitter sp = new SpiderSplitterImpl(opt, log);
		d = sp.splitSpiders(d);

		// Convert diagram to DNF and remove not
		Normaliser n = new NormaliserImpl(opt, log);
		d = n.toDNFWithoutNegation(d);
		
		return d;
	}			

	/** Compute the isShaded flag */
	protected boolean isShaded(Zone z, UnitaryDiagram ud, List zones) {
		// Missing zone
		if (!zones.contains(z)) return true;
		// Shaded zone
		return z.isShaded(ud);
	}

	/** Compute the number of feet */
	protected int computeNoFeet(Zone z, UnitaryDiagram ud) {
		int noFeet = 0;
		Vector spiders = ud.getSpiders();
		for (int i = 0; i < spiders.size(); i++) {
			Spider s = (Spider) spiders.get(i);
			if (s.getHabitat().contains(z))
				noFeet++;
		}
		return noFeet;
	}

	/** Check if there is a contradiction */
	protected boolean isContradiction(boolean isShaded1, int noFeet1, boolean isShaded2, int noFeet2) {
		// Compute result 				
		// z1 is shaded and z2 is shaded
		boolean isCont = false;
		if (isShaded1 && isShaded2) {
			// z1 has exactly n1 elements z2 has exactly n2 elements 
			if (noFeet1 == 0 && noFeet2 == 0) isCont = false;		
			if (noFeet1 == 0 && noFeet2 != 0) isCont = true;		
			if (noFeet1 != 0 && noFeet2 == 0) isCont = true;		
			if (noFeet1 != 0 && noFeet2 != 0) isCont = noFeet1 != noFeet2;		
		// z1 is shaded and z2 is not shaded
		} else if (isShaded1 && !isShaded2) {		
			// z1 has exactly n1 elements z2 has at least n2 elements
			if (noFeet1 == 0 && noFeet2 == 0) isCont = false;		
			if (noFeet1 == 0 && noFeet2 != 0) isCont = true;		
			if (noFeet1 != 0 && noFeet2 == 0) isCont = false;		
			if (noFeet1 != 0 && noFeet2 != 0) isCont = noFeet1 < noFeet2;		
		// z1 is not shaded and z2 is shaded
		} else if (!isShaded1 && isShaded2) {		
			// z1 has at least n1 elements z2 has exactly n2 elements
			if (noFeet1 == 0 && noFeet2 == 0) isCont = false;		
			if (noFeet1 == 0 && noFeet2 != 0) isCont = false;		
			if (noFeet1 != 0 && noFeet2 == 0) isCont = true;		
			if (noFeet1 != 0 && noFeet2 != 0) isCont = noFeet2 < noFeet1;		
		// z1 is not shaded and z2 is not shaded
		} else {
			// z1 has at least n1 elements z2 has at least n2 elements
			isCont = false;
		}
				
		return isCont;
	}
	
	/** Check the type of diagrams */
	private boolean isTrue(Diagram d) {
		return d instanceof TrueDiagram;
	}
	private boolean isFalse(Diagram d) {
		return d instanceof FalseDiagram;
	}
	private boolean isNotTrue(Diagram d) {
		if (d instanceof NotCompoundDiagram) {
			NotCompoundDiagram cd = (NotCompoundDiagram)d;
			Diagram left = (Diagram)cd.getChildren().get(0);
			if (left instanceof TrueDiagram)
				return true;
		}
		return false;
	}
	private boolean isNotFalse(Diagram d) {
		if (d instanceof NotCompoundDiagram) {
			NotCompoundDiagram cd = (NotCompoundDiagram)d;
			Diagram left = (Diagram)cd.getChildren().get(0);
			if (left instanceof FalseDiagram)
				return true;
		}
		return false;
	}
	
	/** Compute all the contours */
	private Set computeContours(UnitaryDiagram d1, UnitaryDiagram d2) {
		Set contours = new TreeSet();
		contours.addAll(d1.getCloneOfContourSet());
		contours.addAll(d2.getCloneOfContourSet());
		return contours;
		
	}

	//
	// Local attributes
	//
	protected Options opt;
	protected ILog log;
}
